#see license.txt for license details
#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/graphics/widgets/eventcal.py
# Event Calendar widget
# author: Andy Robinson

__version__='3.3.0'
__doc__="""This file is a
"""

from reportlab.lib import colors
from reportlab.graphics.shapes import Rect, Drawing, Group, String
from reportlab.graphics.charts.textlabels import Label
from reportlab.graphics.widgetbase import Widget


class EventCalendar(Widget):
    def __init__(self):
        self.x = 0
        self.y = 0
        self.width = 300
        self.height = 150
        self.timeColWidth = None  # if declared, use it; otherwise auto-size.
        self.trackRowHeight = 20
        self.data = []  # list of Event objects
        self.trackNames = None

        self.startTime = None  #displays ALL data on day if not set
        self.endTime = None    # displays ALL data on day if not set
        self.day = 0


        # we will keep any internal geometry variables
        # here.  These are computed by computeSize(),
        # which is the first thing done when drawing.
        self._talksVisible = []  # subset of data which will get plotted, cache
        self._startTime = None
        self._endTime = None
        self._trackCount = 0
        self._colWidths = []
        self._colLeftEdges = []  # left edge of each column

    def computeSize(self):
        "Called at start of draw.  Sets various column widths"
        self._talksVisible = self.getRelevantTalks(self.data)
        self._trackCount = len(self.getAllTracks())
        self.computeStartAndEndTimes()
        self._colLeftEdges = [self.x]
        if self.timeColWidth is None:
            w = self.width / (1 + self._trackCount)
            self._colWidths = [w] * (1+ self._trackCount)
            for i in range(self._trackCount):
                self._colLeftEdges.append(self._colLeftEdges[-1] + w)
        else:
            self._colWidths = [self.timeColWidth]
            w = (self.width - self.timeColWidth) / self._trackCount
            for i in range(self._trackCount):
                self._colWidths.append(w)
                self._colLeftEdges.append(self._colLeftEdges[-1] + w)



    def computeStartAndEndTimes(self):
        "Work out first and last times to display"
        if self.startTime:
            self._startTime = self.startTime
        else:
            for (title, speaker, trackId, day, start, duration) in self._talksVisible:

                if self._startTime is None: #first one
                    self._startTime = start
                else:
                    if start < self._startTime:
                        self._startTime = start

        if self.endTime:
            self._endTime = self.endTime
        else:
            for (title, speaker, trackId, day, start, duration) in self._talksVisible:
                if self._endTime is None: #first one
                    self._endTime = start + duration
                else:
                    if start + duration > self._endTime:
                        self._endTime = start + duration




    def getAllTracks(self):
        tracks = []
        for (title, speaker, trackId, day, hours, duration) in self.data:
            if trackId is not None:
                if trackId not in tracks:
                    tracks.append(trackId)
        tracks.sort()
        return tracks

    def getRelevantTalks(self, talkList):
        "Scans for tracks actually used"
        used = []
        for talk in talkList:
            (title, speaker, trackId, day, hours, duration) = talk
            assert trackId != 0, "trackId must be None or 1,2,3... zero not allowed!"
            if day == self.day:
                if (((self.startTime is None) or ((hours + duration) >= self.startTime))
                and ((self.endTime is None) or (hours <= self.endTime))):
                    used.append(talk)
        return used

    def scaleTime(self, theTime):
        "Return y-value corresponding to times given"
        axisHeight = self.height - self.trackRowHeight
        # compute fraction between 0 and 1, 0 is at start of period
        proportionUp = ((theTime - self._startTime) / (self._endTime - self._startTime))
        y = self.y + axisHeight - (axisHeight * proportionUp)
        return y


    def getTalkRect(self, startTime, duration, trackId, text):
        "Return shapes for a specific talk"
        g = Group()
        y_bottom = self.scaleTime(startTime + duration)
        y_top = self.scaleTime(startTime)
        y_height = y_top - y_bottom

        if trackId is None:
            #spans all columns
            x = self._colLeftEdges[1]
            width = self.width - self._colWidths[0]
        else:
            #trackId is 1-based and these arrays have the margin info in column
            #zero, so no need to add 1
            x = self._colLeftEdges[trackId]
            width = self._colWidths[trackId]

        lab = Label()
        lab.setText(text)
        lab.setOrigin(x + 0.5*width, y_bottom+0.5*y_height)
        lab.boxAnchor = 'c'
        lab.width = width
        lab.height = y_height
        lab.fontSize = 6

        r = Rect(x, y_bottom, width, y_height, fillColor=colors.cyan)
        g.add(r)
        g.add(lab)

        #now for a label
        # would expect to color-code and add text
        return g

    def draw(self):
        self.computeSize()
        g = Group()

        # time column
        g.add(Rect(self.x, self.y, self._colWidths[0], self.height - self.trackRowHeight, fillColor=colors.cornsilk))

        # track headers
        x = self.x + self._colWidths[0]
        y = self.y + self.height - self.trackRowHeight
        for trk in range(self._trackCount):
            wid = self._colWidths[trk+1]
            r = Rect(x, y, wid, self.trackRowHeight, fillColor=colors.yellow)
            s = String(x + 0.5*wid, y, 'Track %d' % trk, align='middle')
            g.add(r)
            g.add(s)
            x = x + wid

        for talk in self._talksVisible:
            (title, speaker, trackId, day, start, duration) = talk
            r = self.getTalkRect(start, duration, trackId, title + '\n' + speaker)
            g.add(r)


        return g




def test():
    "Make a conference event for day 1 of UP Python 2003"


    d = Drawing(400,200)

    cal = EventCalendar()
    cal.x = 50
    cal.y = 25
    cal.data = [
        # these might be better as objects instead of tuples, since I
        # predict a large number of "optionsl" variables to affect
        # formatting in future.

        #title, speaker, track id, day, start time (hrs), duration (hrs)
        # track ID is 1-based not zero-based!
        ('Keynote: Why design another programming language?',  'Guido van Rossum', None, 1, 9.0, 1.0),

        ('Siena Web Service Architecture', 'Marc-Andre Lemburg', 1, 1, 10.5, 1.5),
        ('Extreme Programming in Python', 'Chris Withers', 2, 1, 10.5, 1.5),
        ('Pattern Experiences in C++', 'Mark Radford', 3, 1, 10.5, 1.5),
        ('What is the Type of std::toupper()', 'Gabriel Dos Reis', 4, 1, 10.5, 1.5),
        ('Linguistic Variables: Clear Thinking with Fuzzy Logic ', 'Walter Banks', 5, 1, 10.5, 1.5),

        ('lunch, short presentations, vendor presentations', '', None, 1, 12.0, 2.0),

        ("CORBA? Isn't that obsolete", 'Duncan Grisby', 1, 1, 14.0, 1.5),
        ("Python Design Patterns", 'Duncan Booth', 2, 1, 14.0, 1.5),
        ("Inside Security Checks and Safe Exceptions", 'Brandon Bray', 3, 1, 14.0, 1.5),
        ("Studying at a Distance", 'Panel Discussion, Panel to include Alan Lenton & Francis Glassborow', 4, 1, 14.0, 1.5),
        ("Coding Standards - Given the ANSI C Standard why do I still need a coding Standard", 'Randy Marques', 5, 1, 14.0, 1.5),

        ("RESTful Python", 'Hamish Lawson', 1, 1, 16.0, 1.5),
        ("Parsing made easier - a radical old idea", 'Andrew Koenig', 2, 1, 16.0, 1.5),
        ("C++ & Multimethods", 'Julian Smith', 3, 1, 16.0, 1.5),
        ("C++ Threading", 'Kevlin Henney', 4, 1, 16.0, 1.5),
        ("The Organisation Strikes Back", 'Alan Griffiths & Sarah Lees', 5, 1, 16.0, 1.5),

        ('Birds of a Feather meeting', '', None, 1, 17.5, 2.0),

        ('Keynote: In the Spirit of C',  'Greg Colvin', None, 2, 9.0, 1.0),

        ('The Infinite Filing Cabinet - object storage in Python', 'Jacob Hallen', 1, 2, 10.5, 1.5),
        ('Introduction to Python and Jython for C++ and Java Programmers', 'Alex Martelli', 2, 2, 10.5, 1.5),
        ('Template metaprogramming in Haskell', 'Simon Peyton Jones', 3, 2, 10.5, 1.5),
        ('Plenty People Programming: C++ Programming in a Group, Workshop with a difference', 'Nico Josuttis', 4, 2, 10.5, 1.5),
        ('Design and Implementation of the Boost Graph Library', 'Jeremy Siek', 5, 2, 10.5, 1.5),

        ('lunch, short presentations, vendor presentations', '', None, 2, 12.0, 2.0),

        ("Building GUI Applications with PythonCard and PyCrust", 'Andy Todd', 1, 2, 14.0, 1.5),
        ("Integrating Python, C and C++", 'Duncan Booth', 2, 2, 14.0, 1.5),
        ("Secrets and Pitfalls of Templates", 'Nicolai Josuttis & David Vandevoorde', 3, 2, 14.0, 1.5),
        ("Being a Mentor", 'Panel Discussion, Panel to include Alan Lenton & Francis Glassborow', 4, 2, 14.0, 1.5),
        ("The Embedded C Extensions to C", 'Willem Wakker', 5, 2, 14.0, 1.5),

        ("Lightning Talks", 'Paul Brian', 1, 2, 16.0, 1.5),
        ("Scripting Java Applications with Jython", 'Anthony Eden', 2, 2, 16.0, 1.5),
        ("Metaprogramming and the Boost Metaprogramming Library", 'David Abrahams', 3, 2, 16.0, 1.5),
        ("A Common Vendor ABI for C++ -- GCC's why, what and not", 'Nathan Sidwell & Gabriel Dos Reis', 4, 2, 16.0, 1.5),
        ("The Timing and Cost of Choices", 'Hubert Matthews', 5, 2, 16.0, 1.5),

        ('Birds of a Feather meeting', '', None, 2, 17.5, 2.0),

        ('Keynote: The Cost of C &amp; C++ Compatibility', 'Andy Koenig', None, 3, 9.0, 1.0),

        ('Prying Eyes: Generic Observer Implementations in C++', 'Andrei Alexandrescu', 1, 2, 10.5, 1.5),
        ('The Roadmap to Generative Programming With C++', 'Ulrich Eisenecker', 2, 2, 10.5, 1.5),
        ('Design Patterns in C++ and C# for the Common Language Runtime', 'Brandon Bray', 3, 2, 10.5, 1.5),
        ('Extreme Hour (XH): (workshop) - Jutta Eckstein and Nico Josuttis', 'Jutta Ecstein', 4, 2, 10.5, 1.5),
        ('The Lambda Library : Unnamed Functions for C++', 'Jaako Jarvi', 5, 2, 10.5, 1.5),

        ('lunch, short presentations, vendor presentations', '', None, 3, 12.0, 2.0),

        ('Reflective Metaprogramming', 'Daveed Vandevoorde', 1, 3, 14.0, 1.5),
        ('Advanced Template Issues and Solutions (double session)', 'Herb Sutter',2, 3, 14.0, 3),
        ('Concurrent Programming in Java (double session)', 'Angelika Langer', 3, 3, 14.0, 3),
        ('What can MISRA-C (2nd Edition) do for us?', 'Chris Hills', 4, 3, 14.0, 1.5),
        ('C++ Metaprogramming Concepts and Results', 'Walter E Brown', 5, 3, 14.0, 1.5),

        ('Binding C++ to Python with the Boost Python Library', 'David Abrahams', 1, 3, 16.0, 1.5),
        ('Using Aspect Oriented Programming for Enterprise Application Integration', 'Arno Schmidmeier', 4, 3, 16.0, 1.5),
        ('Defective C++', 'Marc Paterno', 5, 3, 16.0, 1.5),

        ("Speakers' Banquet & Birds of a Feather meeting", '', None, 3, 17.5, 2.0),

        ('Keynote: The Internet, Software and Computers - A Report Card', 'Alan Lenton',  None, 4, 9.0, 1.0),

        ('Multi-Platform Software Development; Lessons from the Boost libraries', 'Beman Dawes', 1, 5, 10.5, 1.5),
        ('The Stability of the C++ ABI', 'Steve Clamage', 2, 5, 10.5, 1.5),
        ('Generic Build Support - A Pragmatic Approach to the Software Build Process', 'Randy Marques', 3, 5, 10.5, 1.5),
        ('How to Handle Project Managers: a survival guide', 'Barb Byro',  4, 5, 10.5, 1.5),

        ('lunch, ACCU AGM', '', None, 5, 12.0, 2.0),

        ('Sauce: An OO recursive descent parser; its design and implementation.', 'Jon Jagger', 1, 5, 14.0, 1.5),
        ('GNIRTS ESAC REWOL -  Bringing the UNIX filters to the C++ iostream library.', 'JC van Winkel', 2, 5, 14.0, 1.5),
        ('Pattern Writing: Live and Direct', 'Frank Buschmann & Kevlin Henney',  3, 5, 14.0, 3.0),
        ('The Future of Programming Languages - A Goldfish Bowl', 'Francis Glassborow and friends',  3, 5, 14.0, 1.5),

        ('Honey, I Shrunk the Threads: Compile-time checked multithreaded transactions in C++', 'Andrei Alexandrescu', 1, 5, 16.0, 1.5),
        ('Fun and Functionality with Functors', 'Lois Goldthwaite', 2, 5, 16.0, 1.5),
        ('Agile Enough?', 'Alan Griffiths', 4, 5, 16.0, 1.5),
        ("Conference Closure: A brief plenary session", '', None, 5, 17.5, 0.5),

        ]

    #return cal
    cal.day = 1

    d.add(cal)


    for format in ['pdf']:#,'gif','png']:
        out = d.asString(format)
        open('eventcal.%s' % format, 'wb').write(out)
        print('saved eventcal.%s' % format)

if __name__=='__main__':
    test()



#
# ReportLab QRCode widget
#
# Ported from the Javascript library QRCode for Javascript by Sam Curren
#
# URL: http://www.d-project.com/
# http://d-project.googlecode.com/svn/trunk/misc/qrcode/js/qrcode.js
# qrcode.js is copyright (c) 2009 Kazuhiko Arase
#
# Original ReportLab module by German M. Bravo
#
# modified and improved by Anders Hammarquist <iko@openend.se>
# and used with permission under the ReportLab License
#
# The word "QR Code" is registered trademark of
# DENSO WAVE INCORPORATED
#   http://www.denso-wave.com/qrcode/faqpatent-e.html

__all__ = ('QrCodeWidget')

import itertools

from reportlab.platypus.flowables import Flowable
from reportlab.graphics.shapes import Group, Rect
from reportlab.lib import colors
from reportlab.lib.validators import isNumber, isNumberOrNone, isColor, Validator
from reportlab.lib.attrmap import AttrMap, AttrMapValue
from reportlab.graphics.widgetbase import Widget
from reportlab.lib.units import mm
from reportlab.lib.utils import asUnicodeEx, isUnicode
from reportlab.graphics.barcode import qrencoder

class isLevel(Validator):
    def test(self, x):
        return x in ['L', 'M', 'Q', 'H']
isLevel = isLevel()

class isUnicodeOrQRList(Validator):
    def _test(self, x):
        if isUnicode(x):
            return True
        if all(isinstance(v, qrencoder.QR) for v in x):
            return True
        return False

    def test(self, x):
        return self._test(x) or self.normalizeTest(x)

    def normalize(self, x):
        if self._test(x):
            return x
        try:
            return asUnicodeEx(x)
        except UnicodeError:
            raise ValueError("Can't convert to unicode: %r" % x)
isUnicodeOrQRList = isUnicodeOrQRList()

class SRect(Rect):
    def __init__(self, x, y, width, height, fillColor=colors.black):
        Rect.__init__(self, x, y, width, height, fillColor=fillColor,
                      strokeColor=None, strokeWidth=0)

class QrCodeWidget(Widget):
    codeName = "QR"
    _attrMap = AttrMap(
        BASE = Widget,
        value = AttrMapValue(isUnicodeOrQRList, desc='QRCode data'),
        x = AttrMapValue(isNumber, desc='x-coord'),
        y = AttrMapValue(isNumber, desc='y-coord'),
        barFillColor = AttrMapValue(isColor, desc='bar color'),
        barWidth = AttrMapValue(isNumber, desc='Width of bars.'), # maybe should be named just width?
        barHeight = AttrMapValue(isNumber, desc='Height of bars.'), # maybe should be named just height?
        barBorder = AttrMapValue(isNumber, desc='Width of QR border.'), # maybe should be named qrBorder?
        barLevel = AttrMapValue(isLevel, desc='QR Code level.'), # maybe should be named qrLevel
        qrVersion = AttrMapValue(isNumberOrNone, desc='QR Code version. None for auto'),
        # Below are ignored, they make no sense
        barStrokeWidth = AttrMapValue(isNumber, desc='Width of bar borders.'),
        barStrokeColor = AttrMapValue(isColor, desc='Color of bar borders.'),
        )
    x = 0
    y = 0
    barFillColor = colors.black
    barStrokeColor = None
    barStrokeWidth = 0
    barHeight = 32*mm
    barWidth = 32*mm
    barBorder = 4
    barLevel = 'L'
    qrVersion = None
    value = None

    def __init__(self, value='Hello World', **kw):
        self.value = isUnicodeOrQRList.normalize(value)
        for k, v in kw.items():
            setattr(self, k, v)

        ec_level = getattr(qrencoder.QRErrorCorrectLevel, self.barLevel)

        self.__dict__['qr'] = qrencoder.QRCode(self.qrVersion, ec_level)

        if isUnicode(self.value):
            self.addData(self.value)
        elif self.value:
            for v in self.value:
                self.addData(v)

    def addData(self, value):
        self.qr.addData(value)

    def draw(self):
        self.qr.make()

        g = Group()

        color = self.barFillColor
        border = self.barBorder
        width = self.barWidth
        height = self.barHeight
        x = self.x
        y = self.y

        g.add(SRect(x, y, width, height, fillColor=None))

        moduleCount = self.qr.getModuleCount()
        minwh = float(min(width, height))
        boxsize = minwh / (moduleCount + border * 2.0)
        offsetX = x + (width - minwh) / 2.0
        offsetY = y + (minwh - height) / 2.0

        for r, row in enumerate(self.qr.modules):
            row = map(bool, row)
            c = 0
            for t, tt in itertools.groupby(row):
                isDark = t
                count = len(list(tt))
                if isDark:
                    x = (c + border) * boxsize
                    y = (r + border + 1) * boxsize
                    s = SRect(offsetX + x, offsetY + height - y, count * boxsize, boxsize,
                            fillColor=color)
                    g.add(s)
                c += count

        return g


# Flowable version

class QrCode(Flowable):
    height = 32*mm
    width = 32*mm
    qrBorder = 4
    qrLevel = 'L'
    qrVersion = None
    value = None

    def __init__(self, value=None, **kw):
        self.value = isUnicodeOrQRList.normalize(value)

        for k, v in kw.items():
            setattr(self, k, v)

        ec_level = getattr(qrencoder.QRErrorCorrectLevel, self.qrLevel)

        self.qr = qrencoder.QRCode(self.qrVersion, ec_level)

        if isUnicode(self.value):
            self.addData(self.value)
        elif self.value:
            for v in self.value:
                self.addData(v)

    def addData(self, value):
        self.qr.addData(value)

    def draw(self):
        self.qr.make()

        moduleCount = self.qr.getModuleCount()
        border = self.qrBorder
        xsize = self.width / (moduleCount + border * 2.0)
        ysize = self.height / (moduleCount + border * 2.0)

        for r, row in enumerate(self.qr.modules):
            row = map(bool, row)
            c = 0
            for t, tt in itertools.groupby(row):
                isDark = t
                count = len(list(tt))
                if isDark:
                    x = (c + border) * xsize
                    y = self.height - (r + border + 1) * ysize
                    self.rect(x, y, count * xsize, ysize * 1.05)
                c += count

    def rect(self, x, y, w, h):
        self.canv.rect(x, y, w, h, stroke=0, fill=1)



#this code is copied/stolen/borrowed/modified from various sources including
#https://github.com/zopefoundation/AccessControl
#https://github.com/zopefoundation/RestrictedPython
#https://github.com/danthedeckie/simpleeval
#hopefully we are standing on giants' shoulders
import sys, os, ast, re, weakref, time, copy, math, types
eval_debug = int(os.environ.get('EVAL_DEBUG','0'))
strTypes = (bytes,str)
isPy39 = sys.version_info[:2]>=(3,9)

haveNameConstant = hasattr(ast,'NameConstant')
import textwrap

class BadCode(ValueError):
	pass

# For AugAssign the operator must be converted to a string.
augOps = {
	ast.Add: '+=',
	ast.Sub: '-=',
	ast.Mult: '*=',
	ast.Div: '/=',
	ast.Mod: '%=',
	ast.Pow: '**=',
	ast.LShift: '<<=',
	ast.RShift: '>>=',
	ast.BitOr: '|=',
	ast.BitXor: '^=',
	ast.BitAnd: '&=',
	ast.FloorDiv: '//=',
	ast.MatMult: '@=',
}

# For creation allowed magic method names. See also
# https://docs.python.org/3/reference/datamodel.html#special-method-names
__allowed_magic_methods__ = frozenset([
	'__init__',
	'__contains__',
	'__lt__',
	'__le__',
	'__eq__',
	'__ne__',
	'__gt__',
	'__ge__',
	])

__rl_unsafe__ = frozenset('''builtins breakpoint __annotations__ co_argcount co_cellvars co_code co_consts
						__code__ co_filename co_firstlineno co_flags co_freevars co_kwonlyargcount
						co_lnotab co_name co_names co_nlocals co_posonlyargcount co_stacksize
						co_varnames cr_await cr_code cr_frame cr_origin cr_running __defaults__
						f_back f_builtins f_code f_exc_traceback f_exc_type f_exc_value f_globals
						f_lasti f_lineno f_locals f_restricted f_trace __func__ func_code func_defaults
						func_doc func_globals func_name gi_code gi_frame gi_running gi_yieldfrom
						__globals__ im_class im_func im_self __iter__ __kwdefaults__ __module__
						__name__ next __qualname__ __self__ tb_frame tb_lasti tb_lineno tb_next
						globals vars locals
						type eval exec aiter anext compile open
						dir print classmethod staticmethod __import__ super property'''.split()
						)
__rl_unsafe_re__ = re.compile(r'\b(?:%s)' % '|'.join(__rl_unsafe__),re.M)


def copy_locations(new_node, old_node):
	ast.copy_location(new_node, old_node)
	ast.fix_missing_locations(new_node)

class UntrustedAstTransformer(ast.NodeTransformer):

	def __init__(self, names_seen=None, nameIsAllowed=None):
		super(UntrustedAstTransformer, self).__init__()
		self.names_seen = {} if names_seen is None else names_seen
		self.nameIsAllowed = nameIsAllowed

		# Global counter to construct temporary variable names.
		self._tmp_idx = 0
		self._tmp_pfx = '_tmp%s' % repr(time.time()).replace('.','')

	@property
	def tmpName(self):
		name = '%s%s' % (self._tmp_pfx,self._tmp_idx)
		self._tmp_idx += 1
		return name

	def error(self, node, msg):
		raise BadCode('Line %s: %s' %  (getattr(node, 'lineno', '??'), msg))

	def guard_iter(self, node):
		"""
		Converts:
			for x in expr
		to
			for x in __rl_getiter__(expr)

		Also used for
		* list comprehensions
		* dict comprehensions
		* set comprehensions
		* generator expresions
		"""
		node = self.visit_children(node)

		if isinstance(node.target, ast.Tuple):
			spec = self.gen_unpack_spec(node.target)
			new_iter = ast.Call(
				func=ast.Name('__rl_iter_unpack_sequence__', ast.Load()),
				args=[node.iter, spec, ast.Name('__rl_getiter__', ast.Load())],
				keywords=[])
		else:
			new_iter = ast.Call(
				func=ast.Name('__rl_getiter__', ast.Load()),
				args=[node.iter],
				keywords=[])

		copy_locations(new_iter, node.iter)
		node.iter = new_iter
		return node

	def is_starred(self, ob):
		return isinstance(ob, ast.Starred)

	def gen_unpack_spec(self, tpl):
		"""Generate a specification for '__rl_unpack_sequence__'.

		This spec is used to protect sequence unpacking.
		The primary goal of this spec is to tell which elements in a sequence
		are sequences again. These 'child' sequences have to be protected
		again.

		For example there is a sequence like this:
			(a, (b, c), (d, (e, f))) = g

		On a higher level the spec says:
			- There is a sequence of len 3
			- The element at index 1 is a sequence again with len 2
			- The element at index 2 is a sequence again with len 2
			  - The element at index 1 in this subsequence is a sequence again
				with len 2

		With this spec '__rl_unpack_sequence__' does something like this for
		protection (len checks are omitted):

			t = list(__rl_getiter__(g))
			t[1] = list(__rl_getiter__(t[1]))
			t[2] = list(__rl_getiter__(t[2]))
			t[2][1] = list(__rl_getiter__(t[2][1]))
			return t

		The 'real' spec for the case above is then:
			spec = {
				'min_len': 3,
				'childs': (
					(1, {'min_len': 2, 'childs': ()}),
					(2, {
							'min_len': 2,
							'childs': (
								(1, {'min_len': 2, 'childs': ()})
							)
						}
					)
				)
			}

		So finally the assignment above is converted into:
			(a, (b, c), (d, (e, f))) = __rl_unpack_sequence__(g, spec)
		"""
		spec = ast.Dict(keys=[], values=[])

		spec.keys.append(ast.Str('childs'))
		spec.values.append(ast.Tuple([], ast.Load()))

		# starred elements in a sequence do not contribute into the min_len.
		# For example a, b, *c = g
		# g must have at least 2 elements, not 3. 'c' is empyt if g has only 2.
		min_len = len([ob for ob in tpl.elts if not self.is_starred(ob)])
		offset = 0

		for idx, val in enumerate(tpl.elts):
			# After a starred element specify the child index from the back.
			# Since it is unknown how many elements from the sequence are
			# consumed by the starred element.
			# For example a, *b, (c, d) = g
			# Then (c, d) has the index '-1'
			if self.is_starred(val):
				offset = min_len + 1

			elif isinstance(val, ast.Tuple):
				el = ast.Tuple([], ast.Load())
				el.elts.append(ast.Num(idx - offset))
				el.elts.append(self.gen_unpack_spec(val))
				spec.values[0].elts.append(el)

		spec.keys.append(ast.Str('min_len'))
		spec.values.append(ast.Num(min_len))

		return spec

	def protect_unpack_sequence(self, target, value):
		spec = self.gen_unpack_spec(target)
		return ast.Call(
			func=ast.Name('__rl_unpack_sequence__', ast.Load()),
			args=[value, spec, ast.Name('__rl_getiter__', ast.Load())],
			keywords=[])

	def gen_unpack_wrapper(self, node, target, ctx='store'):
		"""Helper function to protect tuple unpacks.

		node: used to copy the locations for the new nodes.
		target: is the tuple which must be protected.
		ctx: Defines the context of the returned temporary node.

		It returns a tuple with two element.

		Element 1: Is a temporary name node which must be used to
				   replace the target.
				   The context (store, param) is defined
				   by the 'ctx' parameter..

		Element 2: Is a try .. finally where the body performs the
				   protected tuple unpack of the temporary variable
				   into the original target.
		"""

		# Generate a tmp name to replace the tuple with.
		tnam = self.tmpName

		# Generates an expressions which protects the unpack.
		# converter looks like 'wrapper(tnam)'.
		# 'wrapper' takes care to protect sequence unpacking with __rl_getiter__.
		converter = self.protect_unpack_sequence(
			target,
			ast.Name(tnam, ast.Load()))

		# Assign the expression to the original names.
		# Cleanup the temporary variable.
		# Generates:
		# try:
		#	  # converter is 'wrapper(tnam)'
		#	  arg = converter
		# finally:
		#	  del tmp_arg
		try_body = [ast.Assign(targets=[target], value=converter)]
		finalbody = [self.gen_del_stmt(tnam)]

		cleanup = ast.Try(
			body=try_body, finalbody=finalbody, handlers=[], orelse=[])

		if ctx == 'store':
			ctx = ast.Store()
		elif ctx == 'param':
			ctx = ast.Param()
		else:  # pragma: no cover
			# Only store and param are defined ctx.
			raise NotImplementedError('bad ctx "%s"' % type(ctx))

		# This node is used to catch the tuple in a tmp variable.
		tmp_target = ast.Name(tnam, ctx)

		copy_locations(tmp_target, node)
		copy_locations(cleanup, node)

		return (tmp_target, cleanup)

	def gen_none_node(self):
		return ast.NameConstant(value=None) if hasNameConstant else ast.Name(id='None', ctx=ast.Load())

	def gen_lambda(self, args, body):
		return ast.Lambda(
			args=ast.arguments(
				args=args, vararg=None, kwarg=None, defaults=[]),
			body=body)

	def gen_del_stmt(self, name_to_del):
		return ast.Delete(targets=[ast.Name(name_to_del, ast.Del())])

	def transform_slice(self, slice_):
		"""Transform slices into function parameters.

		ast.Slice nodes are only allowed within a ast.Subscript node.
		To use a slice as an argument of ast.Call it has to be converted.
		Conversion is done by calling the 'slice' function from builtins
		"""

		if isinstance(slice_, ast.Index):
			return slice_.value

		elif isinstance(slice_, ast.Slice):
			# Create a python slice object.
			args = []

			if slice_.lower:
				args.append(slice_.lower)
			else:
				args.append(self.gen_none_node())

			if slice_.upper:
				args.append(slice_.upper)
			else:
				args.append(self.gen_none_node())

			if slice_.step:
				args.append(slice_.step)
			else:
				args.append(self.gen_none_node())

			return ast.Call(
				func=ast.Name('slice', ast.Load()),
				args=args,
				keywords=[])

		elif isinstance(slice_, ast.ExtSlice):
			dims = ast.Tuple([], ast.Load())
			for item in slice_.dims:
				dims.elts.append(self.transform_slice(item))
			return dims

		elif isPy39:
			return slice_

		else:  # pragma: no cover
			# Index, Slice and ExtSlice are only defined Slice types.
			raise NotImplementedError("Unknown slice type: %s" % slice_)

	def isAllowedName(self, node, name):
		if name is None: return
		self.nameIsAllowed(name)

	def check_function_argument_names(self, node):
		# In python3 arguments are always identifiers.
		# In python2 the 'Python.asdl' specifies expressions, but
		# the python grammer allows only identifiers or a tuple of
		# identifiers. If its a tuple 'tuple parameter unpacking' is used,
		# which is gone in python3.
		# See https://www.python.org/dev/peps/pep-3113/

		for arg in node.args.args:
			self.isAllowedName(node, arg.arg)

		if node.args.vararg:
			self.isAllowedName(node, node.args.vararg.arg)

		if node.args.kwarg:
			self.isAllowedName(node, node.args.kwarg.arg)

		for arg in node.args.kwonlyargs:
			self.isAllowedName(node, arg.arg)

	def check_import_names(self, node):
		"""Check the names being imported.

		This is a protection against rebinding dunder names like
		__rl_getitem__,__rl_set__ via imports.

		=> 'from _a import x' is ok, because '_a' is not added to the scope.
		"""
		for name in node.names:
			if '*' in name.name:
				self.error(node, '"*" imports are not allowed.')
			self.isAllowedName(node, name.name)
			if name.asname:
				self.isAllowedName(node, name.asname)

		return self.visit_children(node)

	def gen_attr_check(self, node, attr_name):
		"""Check if 'attr_name' is allowed on the object in node.

		It generates (_getattr_(node, attr_name) and node).
		"""

		call_getattr = ast.Call(
			func=ast.Name('__rl_getattr__', ast.Load()),
			args=[node, ast.Str(attr_name)],
			keywords=[])

		return ast.BoolOp(op=ast.And(), values=[call_getattr, node])

	def visit_Constant(self, node):
		"""Allow constant literals with restriction for Ellipsis.

		Constant replaces Num, Str, Bytes, NameConstant and Ellipsis in
		Python 3.8+.
		:see: https://docs.python.org/dev/whatsnew/3.8.html#deprecated
		"""
		if node.value is Ellipsis:
			# Deny using `...`.
			# Special handling necessary as ``self.not_allowed(node)``
			# would return the Error Message:
			# 'Constant statements are not allowed.'
			# which is only partial true.
			self.error(node, 'Ellipsis statements are not allowed.')
			return
		return self.visit_children(node)

	# ast for Variables
	def visit_Name(self, node):
		node = self.visit_children(node)

		if isinstance(node.ctx, ast.Load):
			if node.id == 'print':
				self.error(node,'print function is not allowed')
			self.names_seen[node.id] = True

		self.isAllowedName(node, node.id)
		return node

	def visit_Call(self, node):
		"""Checks calls with '*args' and '**kwargs'.

		Note: The following happens only if '*args' or '**kwargs' is used.

		Transfroms 'foo(<all the possible ways of args>)' into
		__rl_apply__(foo, <all the possible ways for args>)

		The thing is that '__rl_apply__' has only '*args', '**kwargs', so it gets
		Python to collapse all the myriad ways to call functions
		into one manageable from.

		From there, '__rl_apply__()' wraps args and kws in guarded accessors,
		then calls the function, returning the value.
		"""

		if isinstance(node.func, ast.Name):
			if node.func.id == 'exec':
				self.error(node, 'Exec calls are not allowed.')
			elif node.func.id == 'eval':
				self.error(node, 'Eval calls are not allowed.')

		needs_wrap = False

		for pos_arg in node.args:
			if isinstance(pos_arg, ast.Starred):
				needs_wrap = True

		for keyword_arg in node.keywords:
			if keyword_arg.arg is None:
				needs_wrap = True

		node = self.visit_children(node)

		#if not needs_wrap:
		#	return node

		node.args.insert(0, node.func)
		node.func = ast.Name('__rl_apply__', ast.Load())
		copy_locations(node.func, node.args[0])
		return node

	def visit_Attribute(self, node):
		"""Checks and mutates attribute access/assignment.

		'a.b' becomes '__rl_getattr__(a, "b")'
		"""
		if node.attr.startswith('__') and node.attr != '__':
			self.error(node, '"%s" is an invalid attribute'%node.attr)

		if isinstance(node.ctx, ast.Load):
			node = self.visit_children(node)
			new_node = ast.Call(
				func=ast.Name('__rl_getattr__', ast.Load()),
				args=[node.value, ast.Str(node.attr)],
				keywords=[])

			copy_locations(new_node, node)
			return new_node

		elif isinstance(node.ctx, (ast.Store, ast.Del)):
			node = self.visit_children(node)
			new_value = ast.Call(
				func=ast.Name('__rl_sd__', ast.Load()),
				args=[node.value],
				keywords=[])

			copy_locations(new_value, node.value)
			node.value = new_value
			return node

		else:  # pragma: no cover
			# Impossible Case only ctx Load, Store and Del are defined in ast.
			raise NotImplementedError("Unknown ctx type: %s" % type(node.ctx))

	# Subscripting
	def visit_Subscript(self, node):
		"""Transforms all kinds of subscripts.

		'v[a]' becomes '__rl_getitem__(foo, a)'
		'v[:b]' becomes '__rl_getitem__(foo, slice(None, b, None))'
		'v[a:]' becomes '__rl_getitem__(foo, slice(a, None, None))'
		'v[a:b]' becomes '__rl_getitem__(foo, slice(a, b, None))'
		'v[a:b:c]' becomes '__rl_getitem__(foo, slice(a, b, c))'
		'v[a,b:c] becomes '__rl_getitem__(foo, (a, slice(b, c, None)))'
		#'v[a] = c' becomes '_rl_write__(v)[a] = c'
		#'del v[a]' becomes 'del __rl_sd__(v)[a]'
		"""
		node = self.visit_children(node)

		# 'AugStore' and 'AugLoad' are defined in 'Python.asdl' as possible
		# 'expr_context'. However, according to Python/ast.c
		# they are NOT used by the implementation => No need to worry here.
		# Instead ast.c creates 'AugAssign' nodes, which can be visit_ed.

		if isinstance(node.ctx, ast.Load):
			new_node = ast.Call(
				func=ast.Name('__rl_getitem__', ast.Load()),
				args=[node.value, self.transform_slice(node.slice)],
				keywords=[])

			copy_locations(new_node, node)
			return new_node

		elif isinstance(node.ctx, (ast.Del, ast.Store)):
			#new_value = ast.Call(
			#	func=ast.Name('__rl_sd__', ast.Load()),
			#	args=[node.value],
			#	keywords=[])

			#copy_locations(new_value, node)
			#node.value = new_value
			return node

		else:  # pragma: no cover
			# Impossible Case only ctx Load, Store and Del are defined in ast.
			raise NotImplementedError("Unknown ctx type: %s" % type(node.ctx))

	# Statements
	def visit_Assign(self, node):
		node = self.visit_children(node)

		if not any(isinstance(t, ast.Tuple) for t in node.targets):
			return node

		# Handle sequence unpacking.
		# For briefness this example omits cleanup of the temporary variables.
		# Check 'transform_tuple_assign' how its done.
		#
		# - Single target (with nested support)
		# (a, (b, (c, d))) = <exp>
		# is converted to
		# (a, t1) = __rl_getiter__(<exp>)
		# (b, t2) = __rl_getiter__(t1)
		# (c, d) = __rl_getiter__(t2)
		#
		# - Multi targets
		# (a, b) = (c, d) = <exp>
		# is converted to
		# (c, d) = __rl_getiter__(<exp>)
		# (a, b) = __rl_getiter__(<exp>)
		# Why is this valid ? The original bytecode for this multi targets
		# behaves the same way.

		# ast.NodeTransformer works with list results.
		# He injects it at the rightplace of the node's parent statements.
		new_nodes = []

		# python fills the right most target first.
		for target in reversed(node.targets):
			if isinstance(target, ast.Tuple):
				wrapper = ast.Assign(
					targets=[target],
					value=self.protect_unpack_sequence(target, node.value))
				new_nodes.append(wrapper)
			else:
				new_node = ast.Assign(targets=[target], value=node.value)
				new_nodes.append(new_node)

		for new_node in new_nodes:
			copy_locations(new_node, node)

		return new_nodes

	def visit_AugAssign(self, node):
		"""Forbid certain kinds of AugAssign

		According to the language reference (and ast.c) the following nodes
		are are possible:
		Name, Attribute, Subscript

		Note that although augmented assignment of attributes and
		subscripts is disallowed, augmented assignment of names (such
		as 'n += 1') is allowed.
		'n += 1' becomes 'n = __rl_augAssign__("+=", n, 1)'
		"""

		node = self.visit_children(node)

		if isinstance(node.target, ast.Attribute):
			self.error(node, "Augmented assignment of attributes is not allowed.")

		elif isinstance(node.target, ast.Subscript):
			self.error(node, "Augmented assignment of object items and slices is not allowed.")

		elif isinstance(node.target, ast.Name):
			new_node = ast.Assign(
				targets=[node.target],
				value=ast.Call(
					func=ast.Name('__rl_augAssign__', ast.Load()),
					args=[
						ast.Str(augOps[type(node.op)]),
						ast.Name(node.target.id, ast.Load()),
						node.value
						],
					keywords=[]))

			copy_locations(new_node, node)
			return new_node
		else:  # pragma: no cover
			# Impossible Case - Only Node Types:
			# * Name
			# * Attribute
			# * Subscript
			# defined, those are checked before.
			raise NotImplementedError("Unknown target type: %s" % type(node.target))

	def visit_While(node):
		self.visit_children(node)
		return node

	def visit_ExceptHandler(self, node):
		"""Protect tuple unpacking on exception handlers.

		try:
			.....
		except Exception as (a, b):
			....

		becomes

		try:
			.....
		except Exception as tmp:
			try:
				(a, b) = __rl_getiter__(tmp)
			finally:
				del tmp
		"""
		node = self.visit_children(node)

		self.isAllowedName(node, node.name)
		return node

	def visit_With(self, node):
		"""Protect tuple unpacking on with statements."""
		node = self.visit_children(node)

		items = node.items

		for item in reversed(items):
			if isinstance(item.optional_vars, ast.Tuple):
				tmp_target, unpack = self.gen_unpack_wrapper(
					node,
					item.optional_vars)

				item.optional_vars = tmp_target
				node.body.insert(0, unpack)

		return node

	# Function and class definitions
	def visit_FunctionDef(self, node):
		"""Allow function definitions (`def`) with some restrictions."""
		self.isAllowedName(node, node.name)
		self.check_function_argument_names(node)

		return node

	def visit_Lambda(self, node):
		"""Allow lambda with some restrictions."""
		self.check_function_argument_names(node)

		node = self.visit_children(node)

		return node

	def visit_ClassDef(self, node):
		"""Check the name of a class definition."""
		self.isAllowedName(node, node.name)
		node = self.visit_children(node)
		if any(keyword.arg == 'metaclass' for keyword in node.keywords):
			self.error(node, 'The keyword argument "metaclass" is not allowed.')
		CLASS_DEF = textwrap.dedent('''\
			class %s(metaclass=__metaclass__):
				pass
		''' % node.name)
		new_class_node = ast.parse(CLASS_DEF).body[0]
		new_class_node.body = node.body
		new_class_node.bases = node.bases
		new_class_node.decorator_list = node.decorator_list
		return new_class_node

	# Imports
	def visit_Import(self, node):
		return self.check_import_names(node)

		node = self.visit_children(node)
		new_node = ast.Call(
						func=ast.Name('__rl_add__', ast.Load()),
							args=[node.left, node.right],
							keywords=[])
		copy_locations(new_node, node)
		return new_node

	def visit_BinOp(self,node):
		node = self.visit_children(node)
		op = node.op
		if isinstance(op,(ast.Mult,ast.Add,ast.Pow)):
			opf = ('__rl_mult__' if isinstance(op,ast.Mult)
					else '__rl_add__' if isinstance(op,ast.Add)
					else '__rl_pow__')
			new_node = ast.Call(
						func=ast.Name(opf, ast.Load()),
							args=[node.left, node.right],
							keywords=[])
			copy_locations(new_node, node)
			return new_node
		return node

	visit_ImportFrom = visit_Import
	visit_For = guard_iter
	visit_comprehension = guard_iter

	def generic_visit(self, node):
		"""Reject nodes which do not have a corresponding `visit` method."""
		self.not_allowed(node)

	def not_allowed(self, node):
		self.error(node, '%s statements are not allowed.'%node.__class__.__name__)

	def visit_children(self, node):
		"""Visit the contents of a node."""
		return super(UntrustedAstTransformer, self).generic_visit(node)

	if eval_debug>=2:
		def visit(self, node):
			method = 'visit_' + node.__class__.__name__
			visitor = getattr(self, method, self.generic_visit)
			print('visitor=%s=%r node=%r' % (method,visitor,node))
			return visitor(node)

	visit_Ellipsis = not_allowed
	visit_MatMult = not_allowed
	visit_Exec = not_allowed
	visit_Nonlocal = not_allowed
	visit_AsyncFunctionDef = not_allowed
	visit_Await = not_allowed
	visit_AsyncFor = not_allowed
	visit_AsyncWith = not_allowed
	visit_Print = not_allowed

	visit_Num = visit_children
	visit_Str = visit_children
	visit_Bytes = visit_children
	visit_List = visit_children
	visit_Tuple = visit_children
	visit_Set = visit_children
	visit_Dict = visit_children
	visit_FormattedValue = visit_children
	visit_JoinedStr = visit_children
	visit_NameConstant = visit_children
	visit_Load = visit_children
	visit_Store = visit_children
	visit_Del = visit_children
	visit_Starred = visit_children
	visit_Expression = visit_children
	visit_Expr = visit_children
	visit_UnaryOp = visit_children
	visit_UAdd = visit_children
	visit_USub = visit_children
	visit_Not = visit_children
	visit_Invert = visit_children
	visit_Add = visit_children
	visit_Sub = visit_children
	visit_Mult = visit_children
	visit_Div = visit_children
	visit_FloorDiv = visit_children
	visit_Pow = visit_children
	visit_Mod = visit_children
	visit_LShift = visit_children
	visit_RShift = visit_children
	visit_BitOr = visit_children
	visit_BitXor = visit_children
	visit_BitAnd = visit_children
	visit_BoolOp = visit_children
	visit_And = visit_children
	visit_Or = visit_children
	visit_Compare = visit_children
	visit_Eq = visit_children
	visit_NotEq = visit_children
	visit_Lt = visit_children
	visit_LtE = visit_children
	visit_Gt = visit_children
	visit_GtE = visit_children
	visit_Is = visit_children
	visit_IsNot = visit_children
	visit_In = visit_children
	visit_NotIn = visit_children
	visit_keyword = visit_children
	visit_IfExp = visit_children
	visit_Index = visit_children
	visit_Slice = visit_children
	visit_ExtSlice = visit_children
	visit_ListComp = visit_children
	visit_SetComp = visit_children
	visit_GeneratorExp = visit_children
	visit_DictComp = visit_children
	visit_Raise = visit_children
	visit_Assert = visit_children
	visit_Delete = visit_children
	visit_Pass = visit_children
	visit_alias = visit_children
	visit_If = visit_children
	visit_Break = visit_children
	visit_Continue = visit_children
	visit_Try = visit_children
	visit_TryFinally = visit_children
	visit_TryExcept = visit_children
	visit_withitem = visit_children
	visit_arguments = visit_children
	visit_arg = visit_children
	visit_Return = visit_children
	visit_Yield = visit_children
	visit_YieldFrom = visit_children
	visit_Global = visit_children
	visit_Module = visit_children
	visit_Param = visit_children

def astFormat(node):
	return ast.dump(copy.deepcopy(node),annotate_fields=True, include_attributes=True,indent=4)

class __rl_SafeIter__:
	def __init__(self, it, owner):
		self.__rl_iter__ = owner().__rl_real_iter__(it)
		self.__rl_owner__ = owner

	def __iter__(self):
		return self

	def __next__(self):
		self.__rl_owner__().__rl_check__()
		return	next(self.__rl_iter__)

	next = __next__  # Python 2 compat

__rl_safe_builtins__ = {}	#constructed below
def safer_globals(g=None):
	if g is None:
		g = sys._getframe(1).f_globals.copy()
	for name in ('__annotations__', '__doc__', '__loader__', '__name__', '__package__', '__spec__'):
		if name in g:
			del g[name]
		g['__builtins__'] = __rl_safe_builtins__.copy()
	return g

math_log10 = math.log10
__rl_undef__ = object()
class __RL_SAFE_ENV__:
	__time_time__ = time.time
	__weakref_ref__ = weakref.ref
	__slicetype__ = type(slice(0))
	def __init__(self, timeout=None, allowed_magic_methods=None):
		self.timeout = timeout if timeout is not None else self.__rl_tmax__
		self.allowed_magic_methods = (__allowed_magic_methods__ if allowed_magic_methods==True
									else allowed_magic_methods) if allowed_magic_methods else []
		import builtins
		self.__rl_gen_range__ = builtins.range

		self.__rl_real_iter__ = builtins.iter

		class __rl_dict__(dict):
			def __new__(cls, *args,**kwds):
				if len(args)==1 and not isinstance(args[0],dict):
					try:
						it = self.__real_iter__(args[0])
					except TypeError:
						pass
					else:
						args = (self.__rl_getiter__(it),)
				return dict.__new__(cls,*args,**kwds)

		class __rl_missing_func__:
			def __init__(self,name):
				self.__name__ = name
			def __call__(self,*args,**kwds):
				raise BadCode('missing global %s' % self.__name__)

		self.real_bi = builtins
		self.bi_replace = (
				('open',__rl_missing_func__('open')),
				('iter',self.__rl_getiter__),
				)

		__rl_safe_builtins__.update({_:getattr(builtins,_) for _ in 
			('''None False True abs bool callable chr complex divmod float hash hex id int
		isinstance issubclass len oct ord range repr round slice str tuple setattr
		classmethod staticmethod property divmod next object getattr dict iter pow list
		type max min sum enumerate zip hasattr filter map any all sorted reversed range
		set frozenset

		ArithmeticError AssertionError AttributeError BaseException BufferError BytesWarning
		DeprecationWarning EOFError EnvironmentError Exception FloatingPointError FutureWarning
		GeneratorExit IOError ImportError ImportWarning IndentationError IndexError KeyError
		KeyboardInterrupt LookupError MemoryError NameError NotImplementedError OSError
		OverflowError PendingDeprecationWarning ReferenceError RuntimeError RuntimeWarning
		StopIteration SyntaxError SyntaxWarning SystemError SystemExit TabError TypeError
		UnboundLocalError UnicodeDecodeError UnicodeEncodeError UnicodeError UnicodeTranslateError
		UnicodeWarning UserWarning ValueError Warning ZeroDivisionError
		__build_class__'''
				).split()})

		self.__rl_builtins__ = __rl_builtins__ = {_:__rl_missing_func__(_) for _ in dir(builtins) if callable(getattr(builtins,_))}
		__rl_builtins__.update(__rl_safe_builtins__)

		#these are used in the tree visitor
		__rl_builtins__['__rl_add__'] = self.__rl_add__
		__rl_builtins__['__rl_mult__'] = self.__rl_mult__
		__rl_builtins__['__rl_pow__'] = self.__rl_pow__
		__rl_builtins__['__rl_sd__'] = self.__rl_sd__
		__rl_builtins__['__rl_augAssign__'] = self.__rl_augAssign__
		__rl_builtins__['__rl_getitem__'] = self.__rl_getitem__
		__rl_builtins__['__rl_getattr__'] = self.__rl_getattr__
		__rl_builtins__['__rl_getiter__'] = self.__rl_getiter__
		__rl_builtins__['__rl_max_len__'] = self.__rl_max_len__
		__rl_builtins__['__rl_max_pow_digits__'] = self.__rl_max_pow_digits__
		__rl_builtins__['__rl_iter_unpack_sequence__'] = self.__rl_iter_unpack_sequence__
		__rl_builtins__['__rl_unpack_sequence__'] = self.__rl_unpack_sequence__
		__rl_builtins__['__rl_apply__'] = lambda func,*args,**kwds: self.__rl_apply__(func,args,kwds)
		__rl_builtins__['__rl_SafeIter__'] = __rl_SafeIter__

		#these are tested builtins
		__rl_builtins__['getattr'] = self.__rl_getattr__
		__rl_builtins__['dict'] = __rl_dict__
		__rl_builtins__['iter'] = self.__rl_getiter__
		__rl_builtins__['pow'] = self.__rl_pow__
		__rl_builtins__['list'] = self.__rl_list__
		__rl_builtins__['type'] = self.__rl_type__
		__rl_builtins__['max'] = self.__rl_max__
		__rl_builtins__['min'] = self.__rl_min__
		__rl_builtins__['sum'] = self.__rl_sum__
		__rl_builtins__['enumerate'] = self.__rl_enumerate__
		__rl_builtins__['zip'] = self.__rl_zip__
		__rl_builtins__['hasattr'] = self.__rl_hasattr__
		__rl_builtins__['filter'] = self.__rl_filter__
		__rl_builtins__['map'] = self.__rl_map__
		__rl_builtins__['any'] = self.__rl_any__
		__rl_builtins__['all'] = self.__rl_all__
		__rl_builtins__['sorted'] = self.__rl_sorted__
		__rl_builtins__['reversed'] = self.__rl_reversed__
		__rl_builtins__['range'] = self.__rl_range__
		__rl_builtins__['set'] = self.__rl_set__
		__rl_builtins__['frozenset'] = self.__rl_frozenset__

	def __rl_type__(self,*args):
		if len(args)==1: return type(*args)
		raise BadCode('type call error')

	def __rl_check__(self):
		if self.__time_time__() >= self.__rl_limit__:
			raise BadCode('Resources exceeded')

	def __rl_sd__(self,obj):
		return obj

	def __rl_getiter__(self,it):
		return __rl_SafeIter__(it,owner=self.__weakref_ref__(self))

	def __rl_max__(self,arg,*args,**kwds):
		if args:
			arg = [arg]
			arg.extend(args)
		return max(self.__rl_args_iter__(arg),**kwds)

	def __rl_min__(self,arg,*args,**kwds):
		if args:
			arg = [arg]
			arg.extend(args)
		return min(self.__rl_args_iter__(arg),**kwds)

	def __rl_sum__(self, sequence, start=0):
		return sum(self.__rl_args_iter__(sequence), start)

	def __rl_enumerate__(self, seq):
		return enumerate(self.__rl_args_iter__(seq))

	def __rl_zip__(self,*args):
		return zip(*[self.__rl_args_iter__(self.__rl_getitem__(args, i)) for i in range(len(args))])

	def __rl_hasattr__(self, obj, name):
		try:
			self.__rl_getattr__(obj, name)
		except (AttributeError, BadCode, TypeError):
			return False
		return True

	def __rl_filter__(self, f, seq):
		return filter(f,self.__rl_args_iter__(seq))

	def __rl_map__(self, f, seq):
		return map(f,self.__rl_args_iter__(seq))

	def __rl_any__(self, seq):
		return any(self.__rl_args_iter__(seq))

	def __rl_all__(self, seq):
		return all(self.__rl_args_iter__(seq))

	def __rl_sorted__(self, seq, **kwds):
		return sorted(self.__rl_args_iter__(seq),**kwds)

	def __rl_reversed__(self, seq):
		return self.__rl_args_iter__(reversed(seq))

	def __rl_range__(self,start,*args):
		return self.__rl_getiter__(range(start,*args))

	def __rl_set__(self, it):
		return set(self.__rl_args_iter__(it))

	def __rl_frozenset__(self, it):
		return frozenset(self.__rl_args_iter__(it))

	def __rl_iter_unpack_sequence__(self, it, spec, _getiter_):
		"""Protect sequence unpacking of targets in a 'for loop'.

		The target of a for loop could be a sequence.
		For example "for a, b in it"
		=> Each object from the iterator needs guarded sequence unpacking.
		"""
		# The iteration itself needs to be protected as well.
		for ob in _getiter_(it):
			yield self.__rl_unpack_sequence__(ob, spec, _getiter_)

	def __rl_unpack_sequence__(self, it, spec, _getiter_):
		"""Protect nested sequence unpacking.

		Protect the unpacking of 'it' by wrapping it with '_getiter_'.
		Furthermore for each child element, defined by spec,
		__rl_unpack_sequence__ is called again.

		Have a look at transformer.py 'gen_unpack_spec' for a more detailed
		explanation.
		"""
		# Do the guarded unpacking of the sequence.
		ret = list(self.__rl__getiter__(it))

		# If the sequence is shorter then expected the interpreter will raise
		# 'ValueError: need more than X value to unpack' anyway
		# => No childs are unpacked => nothing to protect.
		if len(ret) < spec['min_len']:
			return ret

		# For all child elements do the guarded unpacking again.
		for (idx, child_spec) in spec['childs']:
			ret[idx] = self.__rl_unpack_sequence__(ret[idx], child_spec, _getiter_)
		return ret

	def __rl_is_allowed_name__(self, name):
		"""Check names if they are allowed.
		If ``allow_magic_methods is True`` names in `__allowed_magic_methods__`
		are additionally allowed although their names start with `_`.
		"""
		if isinstance(name,strTypes):
			if name in __rl_unsafe__ or (name.startswith('__')
				and name!='__'
				and name not in self.allowed_magic_methods):
				raise BadCode('unsafe access of %s' % name)

	def __rl_getattr__(self, obj, a, *args):
		if isinstance(obj, strTypes) and a=='format':
			raise BadCode('%s.format is not implemented' % type(obj))
		self.__rl_is_allowed_name__(a)
		return getattr(obj,a,*args)

	def __rl_getitem__(self, obj, a):
		if type(a) is self.__slicetype__:
			if a.step is not None:
				v = obj[a]
			else:
				start = a.start
				stop = a.stop
				if start is None:
					start = 0
				if stop is None:
					v = obj[start:]
				else:
					v = obj[start:stop]
			return v
		elif isinstance(a,strTypes):
			self.__rl_is_allowed_name__(a)
			return obj[a]
		return obj[a]

	__rl_tmax__ = 5
	__rl_max_len__ = 100000
	__rl_max_pow_digits__ = 100

	def __rl_add__(self, a, b):
		if (hasattr(a, '__len__') and hasattr(b, '__len__')
				and (len(a) + len(b)) > self.__rl_max_len__):
			raise BadCode("excessive length")
		return a + b

	def __rl_mult__(self, a, b):
		if ((hasattr(a, '__len__') and b * len(a) > self.__rl_max_len__)
				or (hasattr(b, '__len__') and a * len(b) > self.__rl_max_len__)):
			raise BadCode("excessive length")
		return a * b

	def __rl_pow__(self, a, b):
		try:
			if b>0:
				if int(b*math_log10(a)+1)>self.__rl_max_pow_digits__:
					raise BadCode
		except:
			raise BadCode('%r**%r invalid or too large' % (a,b))
		return a ** b

	def __rl_augAssign__(self,op,v,i):
		if op=='+=': return self.__rl_add__(v,i)
		if op=='-=': return v-i
		if op=='*=': return self.__rl_mult__(v,i)
		if op=='/=': return v/i
		if op=='%=': return v%i
		if op=='**=': return self.__rl_pow__(v,i)
		if op=='<<=': return v<<i
		if op=='>>=': return v>>i
		if op=='|=': return v|i
		if op=='^=': return v^i
		if op=='&=': return v&i
		if op=='//=': return v//i

	def __rl_apply__(self, func, args, kwds):
		obj = getattr(func,'__self__',None)
		if obj:
			if isinstance(obj,dict) and func.__name__ in ('pop','setdefault','get', 'popitem'):
				self.__rl_is_allowed_name__(args[0])
		return func(*[a for a in self.__rl_getiter__(args)], **{k:v for k,v in kwds.items()})

	def __rl_args_iter__(self,*args):
		if len(args) == 1:
			i = args[0]
			# Don't double-wrap
			if isinstance(i, __rl_SafeIter__):
				return i
			if not isinstance(i,self.__rl_gen_range__):
				return self.__rl_getiter__(i)
		return self.__rl_getiter__(iter(*args))

	def __rl_list__(self,it):
		return list(self.__rl_getiter__(it))

	def __rl_compile__(self, src, fname='<string>', mode="eval", flags=0, inherit=True, visit=None):
		names_seen = {}
		if not visit:
			bcode = compile(src, fname, mode=mode, flags=flags, dont_inherit=not inherit)
		else:
			astc = ast.parse(src, fname, mode)
			if eval_debug>0:
				print('pre:\n%s\n'%astFormat(astc))
			astc = visit(astc)
			if eval_debug>0:
				print('post:\n%s\n'%astFormat(astc))
			bcode = compile(astc, fname, mode=mode)
		return bcode, names_seen

	def __rl_safe_eval__(self, expr, g, l, mode, timeout=None, allowed_magic_methods=None, __frame_depth__=3):
		bcode, ns = self.__rl_compile__(expr, fname='<string>', mode=mode, flags=0, inherit=True,
				visit=UntrustedAstTransformer(nameIsAllowed=self.__rl_is_allowed_name__).visit)
		if None in (l,g):
			G = sys._getframe(__frame_depth__)
			L = G.f_locals.copy() if l is None else l
			G = G.f_globals.copy() if g is None else g
		else:
			G = g
			L = l
		obi = (G['__builtins__'],) if '__builtins__' in G else False
		G['__builtins__'] = self.__rl_builtins__
		self.__rl_limit__ = self.__time_time__() + (timeout if timeout is not None else self.timeout)
		if allowed_magic_methods is not None:
			self.allowed_magic_methods = ( __allowed_magic_methods__ if allowed_magic_methods==True
										else allowed_magic_methods) if allowed_magic_methods else []
		sbi = [].append
		bi = self.real_bi
		bir = self.bi_replace
		for n, r in bir:
			sbi(getattr(bi,n))
			setattr(bi,n,r)
		try:
			return eval(bcode,G,L)
		finally:
			sbi = sbi.__self__
			for i, (n, r) in enumerate(bir):
				setattr(bi,n,sbi[i])
			if obi:
				G['__builtins__'] = obi[0]

class __rl_safe_eval__:
	'''creates one environment and re-uses it'''
	mode = 'eval'
	def __init__(self):
		self.env = None

	def __call__(self, expr, g=None, l=None, timeout=None, allowed_magic_methods=None):
		if not self.env: self.env = __RL_SAFE_ENV__(timeout=timeout, allowed_magic_methods=allowed_magic_methods)
		return self.env.__rl_safe_eval__(expr, g, l, self.mode, timeout=timeout,
			allowed_magic_methods=allowed_magic_methods,
			__frame_depth__=2)

class __rl_safe_exec__(__rl_safe_eval__):
	mode = 'exec'

def rl_extended_literal_eval(expr, safe_callables=None, safe_names=None):
	if safe_callables is None:
		safe_callables = {}
	if safe_names is None:
		safe_names = {}
	safe_names = safe_names.copy()
	safe_names.update({'None': None, 'True': True, 'False': False})
	#make these readonly with MappingProxyType
	safe_names = types.MappingProxyType(safe_names)
	safe_callables = types.MappingProxyType(safe_callables)
	if isinstance(expr, str):
		expr = ast.parse(expr, mode='eval')
	if isinstance(expr, ast.Expression):
		expr = expr.body
	try:
		# Python 3.4 and up
		ast.NameConstant
		safe_test = lambda n: isinstance(n, ast.NameConstant) or isinstance(n,ast.Name) and n.id in safe_names
		safe_extract = lambda n: n.value if isinstance(n,ast.NameConstant) else safe_names[n.id]
	except AttributeError:
		# Everything before
		safe_test = lambda n: isinstance(n, ast.Name) and n.id in safe_names
		safe_extract = lambda n: safe_names[n.id]
	def _convert(node):
		if isinstance(node, (ast.Str, ast.Bytes)):
			return node.s
		elif isinstance(node, ast.Num):
			return node.n
		elif isinstance(node, ast.Tuple):
			return tuple(map(_convert, node.elts))
		elif isinstance(node, ast.List):
			return list(map(_convert, node.elts))
		elif isinstance(node, ast.Dict):
			return dict((_convert(k), _convert(v)) for k, v
						in zip(node.keys, node.values))
		elif safe_test(node):
			return safe_extract(node)
		elif isinstance(node, ast.UnaryOp) and \
			 isinstance(node.op, (ast.UAdd, ast.USub)) and \
			 isinstance(node.operand, (ast.Num, ast.UnaryOp, ast.BinOp)):
			operand = _convert(node.operand)
			if isinstance(node.op, ast.UAdd):
				return + operand
			else:
				return - operand
		elif isinstance(node, ast.BinOp) and \
			 isinstance(node.op, (ast.Add, ast.Sub)) and \
			 isinstance(node.right, (ast.Num, ast.UnaryOp, ast.BinOp)) and \
			 isinstance(node.right.n, complex) and \
			 isinstance(node.left, (ast.Num, ast.UnaryOp, astBinOp)):
			left = _convert(node.left)
			right = _convert(node.right)
			if isinstance(node.op, ast.Add):
				return left + right
			else:
				return left - right
		elif isinstance(node, ast.Call) and \
			 isinstance(node.func, ast.Name) and \
			 node.func.id in safe_callables:
			return safe_callables[node.func.id](
				*[_convert(n) for n in node.args],
				**{kw.arg: _convert(kw.value) for kw in node.keywords})
		raise ValueError('Bad expression')
	return _convert(expr)

rl_safe_exec = __rl_safe_exec__()
rl_safe_eval = __rl_safe_eval__()



#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/graphics/shapes.py

__version__='3.5.60'
__doc__='''Core of the graphics library - defines Drawing and Shapes'''

import os, sys
from math import pi, cos, sin, sqrt, radians, floor

from reportlab.platypus import Flowable
from reportlab.rl_config import shapeChecking, verbose, defaultGraphicsFontName as _baseGFontName, _unset_, decimalSymbol
from reportlab.lib import logger
from reportlab.lib import colors
from reportlab.lib.validators import *
from reportlab.lib.utils import isSeq, asBytes
isOpacity = NoneOr(isNumberInRange(0,1))
from reportlab.lib.attrmap import *
from reportlab.lib.rl_accel import fp_str
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.lib.fonts import tt2ps
from reportlab.pdfgen.canvas import FILL_EVEN_ODD, FILL_NON_ZERO
_baseGFontNameB = tt2ps(_baseGFontName,1,0)
_baseGFontNameI = tt2ps(_baseGFontName,0,1)
_baseGFontNameBI = tt2ps(_baseGFontName,1,1)

# two constants for filling rules
NON_ZERO_WINDING = 'Non-Zero Winding'
EVEN_ODD = 'Even-Odd'

## these can be overridden at module level before you start
#creating shapes.  So, if using a special color model,
#this provides support for the rendering mechanism.
#you can change defaults globally before you start
#making shapes; one use is to substitute another
#color model cleanly throughout the drawing.

STATE_DEFAULTS = {   # sensible defaults for all
    'transform': (1,0,0,1,0,0),

    # styles follow SVG naming
    'strokeColor': colors.black,
    'strokeWidth': 1,
    'strokeLineCap': 0,
    'strokeLineJoin': 0,
    'strokeMiterLimit' : 10,    # don't know yet so let bomb here
    'strokeDashArray': None,
    'strokeOpacity': None, #100%
    'fillOpacity': None,
    'fillOverprint': False,
    'strokeOverprint': False,
    'overprintMask': 0,

    'fillColor': colors.black,   #...or text will be invisible
    'fillMode': FILL_EVEN_ODD,      #same as pdfgen.canvas

    'fontSize': 10,
    'fontName': _baseGFontName,
    'textAnchor':  'start' # can be start, middle, end, inherited
    }

####################################################################
# math utilities.  These are now in reportlab.graphics.transform
####################################################################
from . transform import *

def _textBoxLimits(text, font, fontSize, leading, textAnchor, boxAnchor):
    w = 0
    for t in text:
        w = max(w,stringWidth(t,font, fontSize))

    h = len(text)*leading
    yt = fontSize
    if boxAnchor[0]=='s':
        yb = -h
        yt = yt - h
    elif boxAnchor[0]=='n':
        yb = 0
    else:
        yb = -h/2.0
        yt = yt + yb

    if boxAnchor[-1]=='e':
        xb = -w
        if textAnchor=='end': xt = 0
        elif textAnchor=='start': xt = -w
        else: xt = -w/2.0
    elif boxAnchor[-1]=='w':
        xb = 0
        if textAnchor=='end': xt = w
        elif textAnchor=='start': xt = 0
        else: xt = w/2.0
    else:
        xb = -w/2.0
        if textAnchor=='end': xt = -xb
        elif textAnchor=='start': xt = xb
        else: xt = 0

    return xb, yb, w, h, xt, yt

def _rotatedBoxLimits( x, y, w, h, angle):
    '''
    Find the corner points of the rotated w x h sized box at x,y
    return the corner points and the min max points in the original space
    '''
    C = zTransformPoints(rotate(angle),((x,y),(x+w,y),(x+w,y+h),(x,y+h)))
    X = [x[0] for x in C]
    Y = [x[1] for x in C]
    return min(X), max(X), min(Y), max(Y), C

class _DrawTimeResizeable:
    '''Addin class to provide the horribleness of _drawTimeResize'''
    def _drawTimeResize(self,w,h):
        if hasattr(self,'_canvas'):
            canvas = self._canvas
            drawing = canvas._drawing
            drawing.width, drawing.height = w, h
            if hasattr(canvas,'_drawTimeResize'):
                canvas._drawTimeResize(w,h)

class _SetKeyWordArgs:
    def __init__(self, keywords={}):
        """In general properties may be supplied to the constructor."""
        for key, value in keywords.items():
            setattr(self, key, value)

#################################################################
#
#    Helper functions for working out bounds
#
#################################################################

def getRectsBounds(rectList):
    # filter out any None objects, e.g. empty groups
    L = [x for x in rectList if x is not None]
    if not L: return None

    xMin, yMin, xMax, yMax = L[0]
    for (x1, y1, x2, y2) in L[1:]:
        if x1 < xMin:
            xMin = x1
        if x2 > xMax:
            xMax = x2
        if y1 < yMin:
            yMin = y1
        if y2 > yMax:
            yMax = y2
    return (xMin, yMin, xMax, yMax)

def _getBezierExtrema(y0,y1,y2,y3):
    '''
    this is used to find if a curveTo path operator has extrema in its range
    The curveTo operator is defined by the points y0, y1, y2, y3

        B(t):=(1-t)^3*y0+3*(1-t)^2*t*y1+3*(1-t)*t^2*y2+t^3*y3
            :=t^3*(y3-3*y2+3*y1-y0)+t^2*(3*y2-6*y1+3*y0)+t*(3*y1-3*y0)+y0
    and is a cubic bezier curve.

    The differential is a quadratic
        t^2*(3*y3-9*y2+9*y1-3*y0)+t*(6*y2-12*y1+6*y0)+3*y1-3*y0

    The extrema must be at real roots, r, of the above which lie in 0<=r<=1

    The quadratic coefficients are
        a=3*y3-9*y2+9*y1-3*y0 b=6*y2-12*y1+6*y0 c=3*y1-3*y0
    or
        a=y3-3*y2+3*y1-y0 b=2*y2-4*y1+2*y0 c=y1-y0  (remove common factor of 3)
    or
        a=y3-3*(y2-y1)-y0 b=2*(y2-2*y1+y0) c=y1-y0

    The returned value is [y0,x1,x2,y3] where if found x1, x2 are any extremals that were found;
    there can be 0, 1 or 2 extremals
    '''
    a=y3-3*(y2-y1)-y0
    b=2*(y2-2*y1+y0)
    c=y1-y0
    Y = [y0] #the set of points

    #standard method to find roots of quadratic
    d = b*b - 4*a*c
    if d>=0:
        d = sqrt(d)
        if b<0: d = -d
        q = -0.5*(b+d)
        R = []
        try:
            R.append(q/a)
        except:
            pass
        try:
            R.append(c/q)
        except:
            pass
        b *= 1.5
        c *= 3
        for t in R:
            if 0<=t<=1:
                #real root in range evaluate spline there and add to X
                Y.append(t*(t*(t*a+b)+c)+y0)
    Y.append(y3)
    return Y

def getPathBounds(points):
    n = len(points)
    f = lambda i,p = points: p[i]
    xs = list(map(f,range(0,n,2)))
    ys = list(map(f,range(1,n,2)))
    return (min(xs), min(ys), max(xs), max(ys))

def getPointsBounds(pointList):
    "Helper function for list of points"
    first = pointList[0]
    if isSeq(first):
        xs = [xy[0] for xy in pointList]
        ys = [xy[1] for xy in pointList]
        return (min(xs), min(ys), max(xs), max(ys))
    else:
        return getPathBounds(pointList)

#################################################################
#
#    And now the shapes themselves....
#
#################################################################
class Shape(_SetKeyWordArgs,_DrawTimeResizeable):
    """Base class for all nodes in the tree. Nodes are simply
    packets of data to be created, stored, and ultimately
    rendered - they don't do anything active.  They provide
    convenience methods for verification but do not
    check attribiute assignments or use any clever setattr
    tricks this time."""
    _attrMap = AttrMap()

    def copy(self):
        """Return a clone of this shape."""

        # implement this in the descendants as they need the right init methods.
        raise NotImplementedError("No copy method implemented for %s" % self.__class__.__name__)

    def getProperties(self,recur=1):
        """Interface to make it easy to extract automatic
        documentation"""

        #basic nodes have no children so this is easy.
        #for more complex objects like widgets you
        #may need to override this.
        props = {}
        for key, value in self.__dict__.items():
            if key[0:1] != '_':
                props[key] = value
        return props

    def setProperties(self, props):
        """Supports the bulk setting if properties from,
        for example, a GUI application or a config file."""

        self.__dict__.update(props)
        #self.verify()

    def dumpProperties(self, prefix=""):
        """Convenience. Lists them on standard output.  You
        may provide a prefix - mostly helps to generate code
        samples for documentation."""

        propList = list(self.getProperties().items())
        propList.sort()
        if prefix:
            prefix = prefix + '.'
        for (name, value) in propList:
            print('%s%s = %s' % (prefix, name, value))

    def verify(self):
        """If the programmer has provided the optional
        _attrMap attribute, this checks all expected
        attributes are present; no unwanted attributes
        are present; and (if a checking function is found)
        checks each attribute.  Either succeeds or raises
        an informative exception."""

        if self._attrMap is not None:
            for key in self.__dict__.keys():
                if key[0] != '_':
                    assert key in self._attrMap, "Unexpected attribute %s found in %s" % (key, self)
            for attr, metavalue in self._attrMap.items():
                assert hasattr(self, attr), "Missing attribute %s from %s" % (attr, self)
                value = getattr(self, attr)
                assert metavalue.validate(value), "Invalid value %s for attribute %s in class %s" % (value, attr, self.__class__.__name__)

    if shapeChecking:
        """This adds the ability to check every attribute assignment as it is made.
        It slows down shapes but is a big help when developing. It does not
        get defined if rl_config.shapeChecking = 0"""
        def __setattr__(self, attr, value):
            """By default we verify.  This could be off
            in some parallel base classes."""
            validateSetattr(self,attr,value)    #from reportlab.lib.attrmap

    def getBounds(self):
        "Returns bounding rectangle of object as (x1,y1,x2,y2)"
        raise NotImplementedError("Shapes and widgets must implement getBounds")

class Group(Shape):
    """Groups elements together.  May apply a transform
    to its contents.  Has a publicly accessible property
    'contents' which may be used to iterate over contents.
    In addition, child nodes may be given a name in which
    case they are subsequently accessible as properties."""

    _attrMap = AttrMap(
        transform = AttrMapValue(isTransform,desc="Coordinate transformation to apply",advancedUsage=1),
        contents = AttrMapValue(isListOfShapes,desc="Contained drawable elements"),
        strokeOverprint = AttrMapValue(isBoolean,desc='Turn on stroke overprinting'),
        fillOverprint = AttrMapValue(isBoolean,desc='Turn on fill overprinting',advancedUsage=1),
        overprintMask = AttrMapValue(isBoolean,desc='overprinting for ordinary CMYK',advancedUsage=1),
        )

    def __init__(self, *elements, **keywords):
        """Initial lists of elements may be provided to allow
        compact definitions in literal Python code.  May or
        may not be useful."""

        # Groups need _attrMap to be an instance rather than
        # a class attribute, as it may be extended at run time.
        self._attrMap = self._attrMap.clone()
        self.contents = []
        self.transform = (1,0,0,1,0,0)
        for elt in elements:
            self.add(elt)
        # this just applies keywords; do it at the end so they
        #don;t get overwritten
        _SetKeyWordArgs.__init__(self, keywords)

    def _addNamedNode(self,name,node):
        'if name is not None add an attribute pointing to node and add to the attrMap'
        if name:
            if name not in list(self._attrMap.keys()):
                self._attrMap[name] = AttrMapValue(isValidChild)
            setattr(self, name, node)

    def add(self, node, name=None):
        """Appends non-None child node to the 'contents' attribute. In addition,
        if a name is provided, it is subsequently accessible by name
        """
        # propagates properties down
        if node is not None:
            assert isValidChild(node), "Can only add Shape or UserNode objects to a Group"
            self.contents.append(node)
            self._addNamedNode(name,node)

    def _nn(self,node):
        self.add(node)
        return self.contents[-1]

    def insert(self, i, n, name=None):
        'Inserts sub-node n in contents at specified location'
        if n is not None:
            assert isValidChild(n), "Can only insert Shape or UserNode objects in a Group"
            if i<0:
                self.contents[i:i] =[n]
            else:
                self.contents.insert(i,n)
            self._addNamedNode(name,n)

    def expandUserNodes(self):
        """Return a new object which only contains primitive shapes."""

        # many limitations - shared nodes become multiple ones,
        obj = isinstance(self,Drawing) and Drawing(self.width,self.height) or Group()
        obj._attrMap = self._attrMap.clone()
        if hasattr(obj,'transform'): obj.transform = self.transform[:]

        self_contents = self.contents
        a = obj.contents.append
        for child in self_contents:
            if isinstance(child, UserNode):
                newChild = child.provideNode()
            elif isinstance(child, Group):
                newChild = child.expandUserNodes()
            else:
                newChild = child.copy()
            a(newChild)

        self._copyNamedContents(obj)
        return obj

    def _explode(self):
        ''' return a fully expanded object'''
        obj = Group()
        if hasattr(self,'__label__'):
            obj.__label__=self.__label__
        if hasattr(obj,'transform'): obj.transform = self.transform[:]
        P = self.getContents()[:]   # pending nodes
        while P:
            n = P.pop(0)
            if isinstance(n, UserNode):
                P.append(n.provideNode())
            elif isinstance(n, Group):
                n = n._explode()
                if n.transform==(1,0,0,1,0,0):
                    obj.contents.extend(n.contents)
                else:
                    obj.add(n)
            else:
                obj.add(n)
        return obj

    def _copyContents(self,obj):
        for child in self.contents:
            obj.contents.append(child)

    def _copyNamedContents(self,obj,aKeys=None,noCopy=('contents',)):
        from copy import copy
        self_contents = self.contents
        if not aKeys: aKeys = list(self._attrMap.keys())
        for k, v in self.__dict__.items():
            if v in self_contents:
                pos = self_contents.index(v)
                setattr(obj, k, obj.contents[pos])
            elif k in aKeys and k not in noCopy:
                setattr(obj, k, copy(v))

    def _copy(self,obj):
        """copies to obj"""
        obj._attrMap = self._attrMap.clone()
        self._copyContents(obj)
        self._copyNamedContents(obj)
        return obj

    def copy(self):
        """returns a copy"""
        return self._copy(self.__class__())

    def rotate(self, theta):
        """Convenience to help you set transforms"""
        self.transform = mmult(self.transform, rotate(theta))

    def translate(self, dx, dy):
        """Convenience to help you set transforms"""
        self.transform = mmult(self.transform, translate(dx, dy))

    def scale(self, sx, sy):
        """Convenience to help you set transforms"""
        self.transform = mmult(self.transform, scale(sx, sy))

    def skew(self, kx, ky):
        """Convenience to help you set transforms"""
        self.transform = mmult(mmult(self.transform, skewX(kx)),skewY(ky))

    def shift(self, x, y):
        '''Convenience function to set the origin arbitrarily'''
        self.transform = self.transform[:-2]+(x,y)

    def asDrawing(self, width, height):
        """ Convenience function to make a drawing from a group
            After calling this the instance will be a drawing!
        """
        self.__class__ = Drawing
        self._attrMap.update(self._xtraAttrMap)
        self.width = width
        self.height = height

    def getContents(self):
        '''Return the list of things to be rendered
        override to get more complicated behaviour'''
        b = getattr(self,'background',None)
        C = self.contents
        if b and b not in C: C = [b]+C
        return C

    def getBounds(self):
        if self.contents:
            b = []
            for elem in self.contents:
                b.append(elem.getBounds())
            x1 = getRectsBounds(b)
            if x1 is None: return None
            x1, y1, x2, y2 = x1
            trans = self.transform
            corners = [[x1,y1], [x1, y2], [x2, y1], [x2,y2]]
            newCorners = []
            for corner in corners:
                newCorners.append(transformPoint(trans, corner))
            return getPointsBounds(newCorners)
        else:
            #empty group needs a sane default; this
            #will happen when interactively creating a group
            #nothing has been added to yet.  The alternative is
            #to handle None as an allowed return value everywhere.
            return None

def _addObjImport(obj,I,n=None):
    '''add an import of obj's class to a dictionary of imports''' #'
    from inspect import getmodule
    c = obj.__class__
    m = getmodule(c).__name__
    n = n or c.__name__
    if m not in I:
        I[m] = [n]
    elif n not in I[m]:
        I[m].append(n)

def _repr(self,I=None):
    '''return a repr style string with named fixed args first, then keywords'''
    if isinstance(self,float):
        return fp_str(self)
    elif isSeq(self):
        s = ''
        for v in self:
            s = s + '%s,' % _repr(v,I)
        if isinstance(self,list):
            return '[%s]' % s[:-1]
        else:
            return '(%s%s)' % (s[:-1],len(self)==1 and ',' or '')
    elif self is EmptyClipPath:
        if I: _addObjImport(self,I,'EmptyClipPath')
        return 'EmptyClipPath'
    elif isinstance(self,Shape):
        if I: _addObjImport(self,I)
        from inspect import getfullargspec
        args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations = getfullargspec(self.__init__)
        if defaults:
            kargs = args[-len(defaults):]
            del args[-len(defaults):]
        else:
            kargs = []
        P = self.getProperties()
        s = self.__class__.__name__+'('
        for n in args[1:]:
            v = P.pop(n,None)
            s += '%s,' % _repr(v,I)
        for n in kargs:
            v = P.pop(n,None)
            s += '%s=%s,' % (n,_repr(v,I))
        for n,v in P.items():
            v = P[n]
            s += '%s=%s,' % (n, _repr(v,I))
        return s[:-1]+')'
    else:
        return repr(self)

def _renderGroupPy(G,pfx,I,i=0,indent='\t\t'):
    s = ''
    C = getattr(G,'transform',None)
    if C: s = s + ('%s%s.transform = %s\n' % (indent,pfx,_repr(C)))
    C  = G.contents
    for n in C:
        if isinstance(n, Group):
            npfx = 'v%d' % i
            i += 1
            l = getattr(n,'__label__','')
            if l: l='#'+l
            s = s + '%s%s=%s._nn(Group())%s\n' % (indent,npfx,pfx,l)
            s = s + _renderGroupPy(n,npfx,I,i,indent)
            i -= 1
        else:
            s = s + '%s%s.add(%s)\n' % (indent,pfx,_repr(n,I))
    return s

def _extraKW(self,pfx,**kw):
    kw.update(self.__dict__)
    R = {}
    n = len(pfx)
    for k in kw.keys():
        if k.startswith(pfx):
            R[k[n:]] = kw[k]
    return R

class Drawing(Group, Flowable):
    """Outermost container; the thing a renderer works on.
    This has no properties except a height, width and list
    of contents."""

    _saveModes = {
            'bmp',
            'eps',
            'gif',
            'jpeg',
            'jpg',
            'pct',
            'pdf',
            'pict',
            'png',
            'ps',
            'py',
            'svg',
            'tif',
            'tiff',
            'tiff1',
            'tiffl',
            'tiffp',
            }

    _bmModes = _saveModes - {'eps','pdf','ps','py','svg'}

    _xtraAttrMap = AttrMap(
        width = AttrMapValue(isNumber,desc="Drawing width in points."),
        height = AttrMapValue(isNumber,desc="Drawing height in points."),
        canv = AttrMapValue(None),
        background = AttrMapValue(isValidChildOrNone,desc="Background widget for the drawing e.g. Rect(0,0,width,height)"),
        hAlign = AttrMapValue(OneOf("LEFT", "RIGHT", "CENTER", "CENTRE"), desc="Horizontal alignment within parent document"),
        vAlign = AttrMapValue(OneOf("TOP", "BOTTOM", "CENTER", "CENTRE"), desc="Vertical alignment within parent document"),
        #AR temporary hack to track back up.
        #fontName = AttrMapValue(isStringOrNone),
        renderScale = AttrMapValue(isNumber,desc="Global scaling for rendering"),
        initialFontName = AttrMapValue(isStringOrNone,desc="override the STATE_DEFAULTS value for fontName"),
        initialFontSize = AttrMapValue(isNumberOrNone,desc="override the STATE_DEFAULTS value for fontSize"),
        )

    _attrMap = AttrMap(BASE=Group,
            formats = AttrMapValue(SequenceOf(
                OneOf(*_saveModes),
                lo=1,emptyOK=0), desc='One or more plot modes'),
            )
    _attrMap.update(_xtraAttrMap)

    def __init__(self, width=400, height=200, *nodes, **keywords):
        self.background = None
        Group.__init__(self,*nodes,**keywords)
        self.width = width
        self.height = height
        self.hAlign = 'LEFT'
        self.vAlign = 'BOTTOM'
        self.renderScale = 1.0

    def _renderPy(self):
        I = {
                'reportlab.graphics.shapes': ['_DrawingEditorMixin','Drawing','Group'],
                'reportlab.lib.colors': ['Color','CMYKColor','PCMYKColor'],
            }
        G = _renderGroupPy(self._explode(),'self',I)
        n = 'ExplodedDrawing_' + self.__class__.__name__
        s = '#Autogenerated by ReportLab guiedit do not edit\n'
        for m, o in I.items():
            s = s + 'from %s import %s\n' % (m,str(o)[1:-1].replace("'",""))
        s = s + '\nclass %s(_DrawingEditorMixin,Drawing):\n' % n
        s = s + '\tdef __init__(self,width=%s,height=%s,*args,**kw):\n' % (self.width,self.height)
        s = s + '\t\tDrawing.__init__(self,width,height,*args,**kw)\n'
        s = s + G
        s = s + '\n\nif __name__=="__main__": #NORUNTESTS\n\t%s().save(formats=[\'pdf\'],outDir=\'.\',fnRoot=None)\n' % n
        return s

    def draw(self,showBoundary=_unset_):
        """This is used by the Platypus framework to let the document
        draw itself in a story.  It is specific to PDF and should not
        be used directly."""
        from reportlab.graphics import renderPDF
        renderPDF.draw(self, self.canv, 0, 0,
                showBoundary=showBoundary if showBoundary is not _unset_ else getattr(self,'_showBoundary',_unset_))

    def wrap(self, availWidth, availHeight):
        width = self.width
        height = self.height
        renderScale = self.renderScale
        if renderScale!=1.0:
            width *= renderScale
            height *= renderScale
        return width, height

    def expandUserNodes(self):
        """Return a new drawing which only contains primitive shapes."""
        obj = Group.expandUserNodes(self)
        obj.width = self.width
        obj.height = self.height
        return obj

    def copy(self):
        """Returns a copy"""
        return self._copy(self.__class__(self.width, self.height))

    def asGroup(self,*args,**kw):
        return self._copy(Group(*args,**kw))

    def save(self, formats=None, verbose=None, fnRoot=None, outDir=None, title='', **kw):
        """Saves copies of self in desired location and formats.
        Multiple formats can be supported in one call

        the extra keywords can be of the form
        _renderPM_dpi=96 (which passes dpi=96 to renderPM)
        """
        genFmt = kw.pop('seqNumber','')
        if isinstance(genFmt,int):
            genFmt = '%4d: ' % genFmt
        else:
            genFmt = ''
        genFmt += 'generating %s file %s'
        from reportlab import rl_config
        ext = ''
        if not fnRoot:
            fnRoot = getattr(self,'fileNamePattern',(self.__class__.__name__+'%03d'))
            chartId = getattr(self,'chartId',0)
            if hasattr(chartId,'__call__'):
                chartId = chartId(self)
            if hasattr(fnRoot,'__call__'):
                fnRoot = fnRoot(chartId)
            else:
                try:
                    fnRoot = fnRoot % chartId
                except TypeError as err:
                    #the exact error message changed from 2.2 to 2.3 so we need to
                    #check a substring
                    if str(err).find('not all arguments converted') < 0: raise

        if outDir is None:
            outDir = getattr(self,'outDir',None)
        if hasattr(outDir,'__call__'):
            outDir = outDir(self)
        if os.path.isabs(fnRoot):
            outDir, fnRoot = os.path.split(fnRoot)
        else:
            outDir = outDir or getattr(self,'outDir','.')
        outDir = outDir.rstrip().rstrip(os.sep)
        if not outDir: outDir = '.'
        if not os.path.isabs(outDir): outDir = os.path.join(getattr(self,'_override_CWD',os.path.dirname(sys.argv[0])),outDir)
        if not os.path.isdir(outDir): os.makedirs(outDir)
        fnroot = os.path.normpath(os.path.join(outDir,fnRoot))
        plotMode = os.path.splitext(fnroot)
        if plotMode[1][1:].lower() in self._saveModes:
            fnroot = plotMode[0]

        plotMode = [x.lower() for x in (formats or getattr(self,'formats',['pdf']))]
        verbose = (verbose is not None and (verbose,) or (getattr(self,'verbose',verbose),))[0]
        _saved = logger.warnOnce.enabled, logger.infoOnce.enabled
        logger.warnOnce.enabled = logger.infoOnce.enabled = verbose
        if 'pdf' in plotMode:
            from reportlab.graphics import renderPDF
            filename = fnroot+'.pdf'
            if verbose: print(genFmt % ('PDF',filename))
            renderPDF.drawToFile(self, filename, title, showBoundary=getattr(self,'showBorder',rl_config.showBoundary),**_extraKW(self,'_renderPDF_',**kw))
            ext = ext +  '/.pdf'
            if sys.platform=='mac':
                import macfs, macostools
                macfs.FSSpec(filename).SetCreatorType("CARO", "PDF ")
                macostools.touched(filename)

        for bmFmt in self._bmModes:
            if bmFmt in plotMode:
                from reportlab.graphics import renderPM
                filename = '%s.%s' % (fnroot,bmFmt)
                if verbose: print(genFmt % (bmFmt,filename))
                dtc = getattr(self,'_drawTimeCollector',None)
                if dtc:
                    dtcfmts = getattr(dtc,'formats',[bmFmt])
                    if bmFmt in dtcfmts and not getattr(dtc,'disabled',0):
                        dtc.clear()
                    else:
                        dtc = None
                renderPM.drawToFile(self, filename,fmt=bmFmt,showBoundary=getattr(self,'showBorder',rl_config.showBoundary),**_extraKW(self,'_renderPM_',**kw))
                ext = ext + '/.' + bmFmt
                if dtc: dtc.save(filename)

        if 'eps' in plotMode:
            try:
                from rlextra.graphics import renderPS_SEP as renderPS
            except ImportError:
                from reportlab.graphics import renderPS
            filename = fnroot+'.eps'
            if verbose: print(genFmt % ('EPS',filename))
            renderPS.drawToFile(self,
                                filename,
                                title = fnroot,
                                dept = getattr(self,'EPS_info',['Testing'])[0],
                                company = getattr(self,'EPS_info',['','ReportLab'])[1],
                                preview = getattr(self,'preview',rl_config.eps_preview),
                                showBoundary=getattr(self,'showBorder',rl_config.showBoundary),
                                ttf_embed=getattr(self,'ttf_embed',rl_config.eps_ttf_embed),
                                **_extraKW(self,'_renderPS_',**kw))
            ext = ext +  '/.eps'

        if 'svg' in plotMode:
            from reportlab.graphics import renderSVG
            filename = fnroot+'.svg'
            if verbose: print(genFmt % ('SVG',filename))
            renderSVG.drawToFile(self,
                                filename,
                                showBoundary=getattr(self,'showBorder',rl_config.showBoundary),**_extraKW(self,'_renderSVG_',**kw))
            ext = ext +  '/.svg'

        if 'ps' in plotMode:
            from reportlab.graphics import renderPS
            filename = fnroot+'.ps'
            if verbose: print(genFmt % ('EPS',filename))
            renderPS.drawToFile(self, filename, showBoundary=getattr(self,'showBorder',rl_config.showBoundary),**_extraKW(self,'_renderPS_',**kw))
            ext = ext +  '/.ps'

        if 'py' in plotMode:
            filename = fnroot+'.py'
            if verbose: print(genFmt % ('py',filename))
            with open(filename,'wb') as f:
                f.write(asBytes(self._renderPy().replace('\n',os.linesep)))
            ext = ext +  '/.py'

        logger.warnOnce.enabled, logger.infoOnce.enabled = _saved
        if hasattr(self,'saveLogger'):
            self.saveLogger(fnroot,ext)
        return ext and fnroot+ext[1:] or ''

    def asString(self, format, verbose=None, preview=0, **kw):
        """Converts to an 8 bit string in given format."""
        assert format in self._saveModes, 'Unknown file format "%s"' % format
        from reportlab import rl_config
        #verbose = verbose is not None and (verbose,) or (getattr(self,'verbose',verbose),)[0]
        if format == 'pdf':
            from reportlab.graphics import renderPDF
            return renderPDF.drawToString(self)
        elif format in self._bmModes:
            from reportlab.graphics import renderPM
            return renderPM.drawToString(self, fmt=format,showBoundary=getattr(self,'showBorder',
                            rl_config.showBoundary),**_extraKW(self,'_renderPM_',**kw))
        elif format == 'eps':
            try:
                from rlextra.graphics import renderPS_SEP as renderPS
            except ImportError:
                from reportlab.graphics import renderPS

            return renderPS.drawToString(self,
                                preview = preview,
                                showBoundary=getattr(self,'showBorder',rl_config.showBoundary))
        elif format == 'ps':
            from reportlab.graphics import renderPS
            return renderPS.drawToString(self, showBoundary=getattr(self,'showBorder',rl_config.showBoundary))
        elif format == 'py':
            return self._renderPy()
        elif format == 'svg':
            from reportlab.graphics import renderSVG
            return renderSVG.drawToString(self,showBoundary=getattr(self,'showBorder',rl_config.showBoundary),**_extraKW(self,'_renderSVG_',**kw))

    def resized(self,kind='fit',lpad=0,rpad=0,bpad=0,tpad=0):
        '''return a base class drawing which ensures all the contents fits'''
        C = self.getContents()
        oW = self.width
        oH = self.height
        drawing = Drawing(oW,oH,*C)
        xL,yL,xH,yH = drawing.getBounds()
        if kind=='fit' or (kind=='expand' and (xL<lpad or xH>oW-rpad or yL<bpad or yH>oH-tpad)):
            drawing.width = xH-xL+lpad+rpad
            drawing.height = yH-yL+tpad+bpad
            drawing.transform = (1,0,0,1,lpad-xL,bpad-yL)
        elif kind=='fitx' or (kind=='expandx' and (xL<lpad or xH>oW-rpad)):
            drawing.width = xH-xL+lpad+rpad
            drawing.transform = (1,0,0,1,lpad-xL,0)
        elif kind=='fity' or (kind=='expandy' and (yL<bpad or yH>oH-tpad)):
            drawing.height = yH-yL+tpad+bpad
            drawing.transform = (1,0,0,1,0,bpad-yL)
        return drawing

class _DrawingEditorMixin:
    '''This is a mixin to provide functionality for edited drawings'''
    def _add(self,obj,value,name=None,validate=None,desc=None,pos=None):
        '''
        effectively setattr(obj,name,value), but takes care of things with _attrMaps etc
        '''
        ivc = isValidChild(value)
        if name and hasattr(obj,'_attrMap'):
            if '_attrMap' not in obj.__dict__:
                obj._attrMap = obj._attrMap.clone()
            if ivc and validate is None: validate = isValidChild
            obj._attrMap[name] = AttrMapValue(validate,desc)
        if hasattr(obj,'add') and ivc:
            if pos:
                obj.insert(pos,value,name)
            else:
                obj.add(value,name)
        elif name:
            setattr(obj,name,value)
        else:
            raise ValueError("Can't add, need name")

class isStrokeDashArray(Validator):
    def test(self,x):
        return isListOfNumbersOrNone.test(x) or (isinstance(x,(list,tuple)) and isNumber(x[0]) and isListOfNumbers(x[1]))
isStrokeDashArray = isStrokeDashArray()

class LineShape(Shape):
    # base for types of lines

    _attrMap = AttrMap(
        strokeColor = AttrMapValue(isColorOrNone),
        strokeWidth = AttrMapValue(isNumber),
        strokeLineCap = AttrMapValue(OneOf(0,1,2),desc="Line cap 0=butt, 1=round & 2=square"),
        strokeLineJoin = AttrMapValue(OneOf(0,1,2),desc="Line join 0=miter, 1=round & 2=bevel"),
        strokeMiterLimit = AttrMapValue(isNumber,desc="miter limit control miter line joins"),
        strokeDashArray = AttrMapValue(isStrokeDashArray,desc="[numbers] or [phase,[numbers]]"),
        strokeOpacity = AttrMapValue(isOpacity,desc="The level of transparency of the line, any real number betwen 0 and 1"),
        strokeOverprint = AttrMapValue(isBoolean,desc='Turn on stroke overprinting'),
        overprintMask = AttrMapValue(isBoolean,desc='overprinting for ordinary CMYK',advancedUsage=1),
        )

    def __init__(self, kw):
        self.strokeColor = STATE_DEFAULTS['strokeColor']
        self.strokeWidth = 1
        self.strokeLineCap = 0
        self.strokeLineJoin = 0
        self.strokeMiterLimit = 0
        self.strokeDashArray = None
        self.strokeOpacity = None
        self.setProperties(kw)

class Line(LineShape):
    _attrMap = AttrMap(BASE=LineShape,
        x1 = AttrMapValue(isNumber,desc=""),
        y1 = AttrMapValue(isNumber,desc=""),
        x2 = AttrMapValue(isNumber,desc=""),
        y2 = AttrMapValue(isNumber,desc=""),
        )

    def __init__(self, x1, y1, x2, y2, **kw):
        LineShape.__init__(self, kw)
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    def getBounds(self):
        "Returns bounding rectangle of object as (x1,y1,x2,y2)"
        return (self.x1, self.y1, self.x2, self.y2)

class SolidShape(LineShape):
    # base for anything with outline and content

    _attrMap = AttrMap(BASE=LineShape,
        fillColor = AttrMapValue(isColorOrNone,desc="filling color of the shape, e.g. red"),
        fillOpacity = AttrMapValue(isOpacity,desc="the level of transparency of the color, any real number between 0 and 1"),
        fillOverprint = AttrMapValue(isBoolean,desc='Turn on fill overprinting'),
        overprintMask = AttrMapValue(isBoolean,desc='overprinting for ordinary CMYK',advancedUsage=1),
        fillMode = AttrMapValue(OneOf(FILL_EVEN_ODD,FILL_NON_ZERO)),
        )

    def __init__(self, kw):
        self.fillColor = STATE_DEFAULTS['fillColor']
        self.fillOpacity = None
        # do this at the end so keywords overwrite
        #the above settings
        LineShape.__init__(self, kw)

# path operator  constants
_MOVETO, _LINETO, _CURVETO, _CLOSEPATH = list(range(4))
_PATH_OP_ARG_COUNT = (2, 2, 6, 0)  # [moveTo, lineTo, curveTo, closePath]
_PATH_OP_NAMES=['moveTo','lineTo','curveTo','closePath']

def _renderPath(path,drawFuncs,countOnly=False,forceClose=False):
    """Helper function for renderers."""
    # this could be a method of Path...
    points = path.points
    i = 0
    hadClosePath = 0
    hadMoveTo = 0
    active = not countOnly
    for op in path.operators:
        if op == _MOVETO:
            if forceClose:
                if hadMoveTo and pop!=_CLOSEPATH:
                    hadClosePath += 1
                    if active:
                        drawFuncs[_CLOSEPATH]()
            hadMoveTo += 1
        nArgs = _PATH_OP_ARG_COUNT[op]
        j = i + nArgs
        drawFuncs[op](*points[i:j])
        i = j
        if op == _CLOSEPATH:
            hadClosePath += 1
        pop = op
    if forceClose and hadMoveTo and pop!=_CLOSEPATH:
        hadClosePath += 1
        if active:
            drawFuncs[_CLOSEPATH]()
    return hadMoveTo == hadClosePath

_fillModeMap = {
        None: None,
        FILL_NON_ZERO: FILL_NON_ZERO,
        'non-zero': FILL_NON_ZERO,
        'nonzero': FILL_NON_ZERO,
        FILL_EVEN_ODD: FILL_EVEN_ODD,
        'even-odd': FILL_EVEN_ODD,
        'evenodd': FILL_EVEN_ODD,
        }

class Path(SolidShape):
    """Path, made up of straight lines and bezier curves."""

    _attrMap = AttrMap(BASE=SolidShape,
        points = AttrMapValue(isListOfNumbers),
        operators = AttrMapValue(isListOfNumbers),
        isClipPath = AttrMapValue(isBoolean),
        autoclose = AttrMapValue(NoneOr(OneOf('svg','pdf'))),
        fillMode = AttrMapValue(OneOf(FILL_EVEN_ODD,FILL_NON_ZERO)),
        )

    def __init__(self, points=None, operators=None, isClipPath=0, autoclose=None, fillMode=FILL_EVEN_ODD, **kw):
        SolidShape.__init__(self, kw)
        if points is None:
            points = []
        if operators is None:
            operators = []
        assert len(points) % 2 == 0, 'Point list must have even number of elements!'
        self.points = points
        self.operators = operators
        self.isClipPath = isClipPath
        self.autoclose=autoclose
        self.fillMode = fillMode

    def copy(self):
        new = self.__class__(self.points[:], self.operators[:])
        new.setProperties(self.getProperties())
        return new

    def moveTo(self, x, y):
        self.points.extend([x, y])
        self.operators.append(_MOVETO)

    def lineTo(self, x, y):
        self.points.extend([x, y])
        self.operators.append(_LINETO)

    def curveTo(self, x1, y1, x2, y2, x3, y3):
        self.points.extend([x1, y1, x2, y2, x3, y3])
        self.operators.append(_CURVETO)

    def closePath(self):
        self.operators.append(_CLOSEPATH)

    def getBounds(self):
        points = self.points
        try:    #in case this complex algorithm is not yet ready :)
            X = []
            aX = X.append
            eX = X.extend
            Y=[]
            aY = Y.append
            eY = Y.extend
            i = 0
            for op in self.operators:
                nArgs = _PATH_OP_ARG_COUNT[op]
                j = i + nArgs
                if nArgs==2:
                    #either moveTo or lineT0
                    aX(points[i])
                    aY(points[i+1])
                elif nArgs==6:
                    #curveTo
                    x1,x2,x3 = points[i:j:2]
                    eX(_getBezierExtrema(X[-1],x1,x2,x3))
                    y1,y2,y3 = points[i+1:j:2]
                    eY(_getBezierExtrema(Y[-1],y1,y2,y3))
                i = j
            return min(X),min(Y),max(X),max(Y)
        except:
            return getPathBounds(points)

EmptyClipPath=Path()    #special path

def getArcPoints(centerx, centery, radius, startangledegrees, endangledegrees, yradius=None, degreedelta=None, reverse=None):
    if yradius is None: yradius = radius
    points = []
    degreestoradians = pi/180.0
    startangle = startangledegrees*degreestoradians
    endangle = endangledegrees*degreestoradians
    while endangle<startangle:
        endangle = endangle+2*pi
    angle = float(endangle - startangle)
    a = points.append
    if angle>.001:
        degreedelta = min(angle,degreedelta or 1.)
        radiansdelta = degreedelta*degreestoradians
        n = max(int(angle/radiansdelta+0.5),1)
        radiansdelta = angle/n
        n += 1
    else:
        n = 1
        radiansdelta = 0

    for angle in range(n):
        angle = startangle+angle*radiansdelta
        a((centerx+radius*cos(angle),centery+yradius*sin(angle)))

    if reverse: points.reverse()
    return points

class ArcPath(Path):
    '''Path with an addArc method'''
    def addArc(self, centerx, centery, radius, startangledegrees, endangledegrees, yradius=None, degreedelta=None, moveTo=None, reverse=None):
        P = getArcPoints(centerx, centery, radius, startangledegrees, endangledegrees, yradius=yradius, degreedelta=degreedelta, reverse=reverse)
        if moveTo or not len(self.operators):
            self.moveTo(P[0][0],P[0][1])
            del P[0]
        for x, y in P: self.lineTo(x,y)

def definePath(pathSegs=[],isClipPath=0, dx=0, dy=0, **kw):
    O = []
    P = []
    for seg in pathSegs:
        if not isSeq(seg):
            opName = seg
            args = []
        else:
            opName = seg[0]
            args = seg[1:]
        if opName not in _PATH_OP_NAMES:
            raise ValueError('bad operator name %s' % opName)
        op = _PATH_OP_NAMES.index(opName)
        if len(args)!=_PATH_OP_ARG_COUNT[op]:
            raise ValueError('%s bad arguments %s' % (opName,str(args)))
        O.append(op)
        P.extend(list(args))
    for d,o in (dx,0), (dy,1):
        for i in range(o,len(P),2):
            P[i] = P[i]+d

    #if there's a bounding box given we constrain so our points lie in it
    #partial bbox is allowed and does something sensible
    bbox = kw.pop('bbox',None)
    if bbox:
        for j in 0,1:
            d = bbox[j],bbox[j+2]
            if d[0] is None and d[1] is None: continue
            a = P[j::2]
            a, b = min(a), max(a)
            if d[0] is not None and d[1] is not None:
                c, d = min(d), max(d)
                fac = (b-a)
                if abs(fac)>=1e-6:
                    fac = (d-c)/fac
                    for i in range(j,len(P),2):
                        P[i] = c + fac*(P[i]-a)
                else:
                    #there's no range in the bbox so fixed as average
                    for i in range(j,len(P),2):
                        P[i] = (c + d)*0.5
            else:
                #if   there's a  lower bound shift so min is lower bound
                #else there's an upper bound shift so max is upper bound
                c = d[0] - a if d[0] is not None else d[1] - b
                for i in range(j,len(P),2):
                    P[i] += c

    return Path(P,O,isClipPath,**kw)

class Rect(SolidShape):
    """Rectangle, possibly with rounded corners."""

    _attrMap = AttrMap(BASE=SolidShape,
        x = AttrMapValue(isNumber),
        y = AttrMapValue(isNumber),
        width = AttrMapValue(isNumber,desc="width of the object in points"),
        height = AttrMapValue(isNumber,desc="height of the objects in points"),
        rx = AttrMapValue(isNumber),
        ry = AttrMapValue(isNumber),
        )

    def __init__(self, x, y, width, height, rx=0, ry=0, **kw):
        SolidShape.__init__(self, kw)
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.rx = rx
        self.ry = ry

    def copy(self):
        new = self.__class__(self.x, self.y, self.width, self.height)
        new.setProperties(self.getProperties())
        return new

    def getBounds(self):
        return (self.x, self.y, self.x + self.width, self.y + self.height)

class Image(SolidShape):
    """Bitmap image."""

    _attrMap = AttrMap(BASE=SolidShape,
        x = AttrMapValue(isNumber),
        y = AttrMapValue(isNumber),
        width = AttrMapValue(isNumberOrNone,desc="width of the object in points"),
        height = AttrMapValue(isNumberOrNone,desc="height of the objects in points"),
        path = AttrMapValue(None),
        )

    def __init__(self, x, y, width, height, path, **kw):
        SolidShape.__init__(self, kw)
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.path = path

    def copy(self):
        new = self.__class__(self.x, self.y, self.width, self.height, self.path)
        new.setProperties(self.getProperties())
        return new

    def getBounds(self):
        # bug fix contributed by Marcel Tromp <mtromp.docbook@gmail.com>
        return (self.x, self.y, self.x + self.width, self.y + self.height)

class Circle(SolidShape):

    _attrMap = AttrMap(BASE=SolidShape,
        cx = AttrMapValue(isNumber,desc="x of the centre"),
        cy = AttrMapValue(isNumber,desc="y of the centre"),
        r = AttrMapValue(isNumber,desc="radius in points"),
        )

    def __init__(self, cx, cy, r, **kw):
        SolidShape.__init__(self, kw)
        self.cx = cx
        self.cy = cy
        self.r = r

    def copy(self):
        new = self.__class__(self.cx, self.cy, self.r)
        new.setProperties(self.getProperties())
        return new

    def getBounds(self):
        return (self.cx - self.r, self.cy - self.r, self.cx + self.r, self.cy + self.r)

class Ellipse(SolidShape):
    _attrMap = AttrMap(BASE=SolidShape,
        cx = AttrMapValue(isNumber,desc="x of the centre"),
        cy = AttrMapValue(isNumber,desc="y of the centre"),
        rx = AttrMapValue(isNumber,desc="x radius"),
        ry = AttrMapValue(isNumber,desc="y radius"),
        )

    def __init__(self, cx, cy, rx, ry, **kw):
        SolidShape.__init__(self, kw)
        self.cx = cx
        self.cy = cy
        self.rx = rx
        self.ry = ry

    def copy(self):
        new = self.__class__(self.cx, self.cy, self.rx, self.ry)
        new.setProperties(self.getProperties())
        return new

    def getBounds(self):
            return (self.cx - self.rx, self.cy - self.ry, self.cx + self.rx, self.cy + self.ry)

class Wedge(SolidShape):
    """A "slice of a pie" by default translates to a polygon moves anticlockwise
       from start angle to end angle"""

    _attrMap = AttrMap(BASE=SolidShape,
        centerx = AttrMapValue(isNumber,desc="x of the centre"),
        centery = AttrMapValue(isNumber,desc="y of the centre"),
        radius = AttrMapValue(isNumber,desc="radius in points"),
        startangledegrees = AttrMapValue(isNumber),
        endangledegrees = AttrMapValue(isNumber),
        yradius = AttrMapValue(isNumberOrNone),
        radius1 = AttrMapValue(isNumberOrNone),
        yradius1 = AttrMapValue(isNumberOrNone),
        annular = AttrMapValue(isBoolean,desc='treat as annular ring'),
        )

    degreedelta = 1 # jump every 1 degrees

    def __init__(self, centerx, centery, radius, startangledegrees, endangledegrees, yradius=None,
            annular=False, **kw):
        SolidShape.__init__(self, kw)
        while endangledegrees<startangledegrees:
            endangledegrees = endangledegrees+360
        #print "__init__"
        self.centerx, self.centery, self.radius, self.startangledegrees, self.endangledegrees = \
            centerx, centery, radius, startangledegrees, endangledegrees
        self.yradius = yradius
        self.annular = annular

    def _xtraRadii(self):
        yradius = getattr(self, 'yradius', None)
        if yradius is None: yradius = self.radius
        radius1 = getattr(self,'radius1', None)
        yradius1 = getattr(self,'yradius1',radius1)
        if radius1 is None: radius1 = yradius1
        return yradius, radius1, yradius1

    #def __repr__(self):
    #        return "Wedge"+repr((self.centerx, self.centery, self.radius, self.startangledegrees, self.endangledegrees ))
    #__str__ = __repr__

    def asPolygon(self):
        #print "asPolygon"
        centerx= self.centerx
        centery = self.centery
        radius = self.radius
        yradius, radius1, yradius1 = self._xtraRadii()
        startangledegrees = self.startangledegrees
        endangledegrees = self.endangledegrees
        degreestoradians = pi/180.0
        startangle = startangledegrees*degreestoradians
        endangle = endangledegrees*degreestoradians
        while endangle<startangle:
            endangle = endangle+2*pi
        angle = float(endangle-startangle)
        points = []
        if angle>0.001:
            degreedelta = min(self.degreedelta or 1.,angle)
            radiansdelta = degreedelta*degreestoradians
            n = max(1,int(angle/radiansdelta+0.5))
            radiansdelta = angle/n
            n += 1
        else:
            n = 1
            radiansdelta = 0
        CA = []
        CAA = CA.append
        a = points.append
        for angle in range(n):
            angle = startangle+angle*radiansdelta
            CAA((cos(angle),sin(angle)))
        for c,s in CA:
            a(centerx+radius*c)
            a(centery+yradius*s)
        if (radius1==0 or radius1 is None) and (yradius1==0 or yradius1 is None):
            a(centerx); a(centery)
        else:
            CA.reverse()
            for c,s in CA:
                a(centerx+radius1*c)
                a(centery+yradius1*s)
        if self.annular:
            P = Path(fillMode=getattr(self,'fillMode', FILL_EVEN_ODD))
            P.moveTo(points[0],points[1])
            for x in range(2,2*n,2):
                P.lineTo(points[x],points[x+1])
            P.closePath()
            P.moveTo(points[2*n],points[2*n+1])
            for x in range(2*n+2,4*n,2):
                P.lineTo(points[x],points[x+1])
            P.closePath()
            return P
        else:
            return Polygon(points)

    def copy(self):
        new = self.__class__(self.centerx,
                    self.centery,
                    self.radius,
                    self.startangledegrees,
                    self.endangledegrees)
        new.setProperties(self.getProperties())
        return new

    def getBounds(self):
        return self.asPolygon().getBounds()

class Polygon(SolidShape):
    """Defines a closed shape; Is implicitly
    joined back to the start for you."""

    _attrMap = AttrMap(BASE=SolidShape,
        points = AttrMapValue(isListOfNumbers,desc="list of numbers in the form x1, y1, x2, y2 ... xn, yn"),
        )

    def __init__(self, points=[], **kw):
        SolidShape.__init__(self, kw)
        assert len(points) % 2 == 0, 'Point list must have even number of elements!'
        self.points = points or []

    def copy(self):
        new = self.__class__(self.points)
        new.setProperties(self.getProperties())
        return new

    def getBounds(self):
        return getPointsBounds(self.points)

class PolyLine(LineShape):
    """Series of line segments.  Does not define a
    closed shape; never filled even if apparently joined.
    Put the numbers in the list, not two-tuples."""

    _attrMap = AttrMap(BASE=LineShape,
        points = AttrMapValue(isListOfNumbers,desc="list of numbers in the form x1, y1, x2, y2 ... xn, yn"),
        )

    def __init__(self, points=[], **kw):
        LineShape.__init__(self, kw)
        points = points or []
        lenPoints = len(points)
        if lenPoints:
            if isSeq(points[0]):
                L = []
                for (x,y) in points:
                    L.append(x)
                    L.append(y)
                points = L
            else:
                assert len(points) % 2 == 0, 'Point list must have even number of elements!'
        self.points = points

    def copy(self):
        new = self.__class__(self.points)
        new.setProperties(self.getProperties())
        return new

    def getBounds(self):
        return getPointsBounds(self.points)

class Hatching(Path):
    '''define a hatching of a set of polygons defined by lists of the form [x0,y0,x1,y1,....,xn,yn]'''

    _attrMap = AttrMap(BASE=Path,
        xyLists = AttrMapValue(EitherOr((isListOfNumbers,SequenceOf(isListOfNumbers,lo=1)),"xy list(s)"),desc="list(s) of numbers in the form x1, y1, x2, y2 ... xn, yn"),
        angles = AttrMapValue(EitherOr((isNumber,isListOfNumbers,"angle(s)")),desc="the angle or list of angles at which hatching lines should be drawn"),
        spacings = AttrMapValue(EitherOr((isNumber,isListOfNumbers,"spacings(s)")),desc="orthogonal distance(s) between hatching lines"),
        )

    def __init__(self, spacings=2, angles=45, xyLists=[], **kwds):
        Path.__init__(self, **kwds)
        if isListOfNumbers(xyLists):
            xyLists  = (xyLists,)
        if isNumber(angles):
            angles = (angles,)  #turn into a sequence
        if isNumber(spacings):
            spacings = (spacings,)  #turn into a sequence
        i = len(angles)-len(spacings)
        if i>0:
            spacings = list(spacings)+i*[spacings[-1]]

        self.xyLists = xyLists
        self.angles = angles
        self.spacings = spacings

        moveTo = self.moveTo
        lineTo = self.lineTo

        for i, theta in enumerate(angles):
            spacing = spacings[i]
            theta = radians(theta)
            cosTheta = cos(theta)
            sinTheta = sin(theta)

            spanMin = 0x7fffffff
            spanMax = -spanMin

            #   Loop to determine the span over which diagonal lines must be drawn.
            for P in xyLists:
                for j in range(0,len(P),2):
                    #   rotated point, since the stripes may be at an angle.
                    y = P[j+1]*cosTheta-P[j]*sinTheta
                    spanMin = min(y,spanMin)
                    spanMax = max(y,spanMax)

            #   Turn the span into a discrete step range.
            spanStart = int(floor(spanMin/spacing)) - 1
            spanEnd  = int(floor(spanMax/spacing)) + 1

            #   Loop to create all stripes.
            for step in range(spanStart,spanEnd):
                nodeX = []
                stripeY = spacing*step

                #   Loop to build a node list for one row of stripes.
                for P in xyLists:
                    k = len(P)-2    #start by comparing with the last point
                    for j in range(0,len(P),2):
                        a = P[k]
                        b = P[k+1]
                        a1 = a*cosTheta + b*sinTheta
                        b1 = b*cosTheta - a*sinTheta
                        x = P[j]
                        y = P[j+1]

                        x1 = x*cosTheta+y*sinTheta
                        y1 = y*cosTheta-x*sinTheta

                        #   Find the node, if any.
                        if (b1<stripeY and y1>=stripeY) or y1<stripeY and b1>=stripeY:
                            nodeX.append(a1+(x1-a1)*(stripeY-b1)/(y1-b1))

                        k = j

                nodeX.sort()

                #   Loop to draw one row of line segments.
                for j in range(0,len(nodeX),2):
                    #   Rotate the points back to their original coordinate system.
                    a = nodeX[j]*cosTheta - stripeY*sinTheta
                    b = stripeY*cosTheta+nodeX[j]*sinTheta
                    x = nodeX[j+1]*cosTheta - stripeY*sinTheta
                    y = stripeY*cosTheta + nodeX[j+1]*sinTheta

                    #Draw a single stripe segment.
                    moveTo(a,b)
                    lineTo(x,y)

def numericXShift(tA,text,w,fontName,fontSize,encoding=None,pivotCharacter=decimalSymbol):
    dp = getattr(tA,'_dp',pivotCharacter)
    i = text.rfind(dp)
    if i>=0:
        dpOffs = getattr(tA,'_dpLen',0)
        w = dpOffs + stringWidth(text[:i],fontName,fontSize,encoding)
    return w

class String(Shape):
    """Not checked against the spec, just a way to make something work.
    Can be anchored left, middle or end."""

    # to do.
    _attrMap = AttrMap(
        x = AttrMapValue(isNumber,desc="x point of anchoring"),
        y = AttrMapValue(isNumber,desc="y point of anchoring"),
        text = AttrMapValue(isString,desc="the text of the string"),
        fontName = AttrMapValue(None,desc="font name of the text - font is either acrobat standard or registered when using external font."),
        fontSize = AttrMapValue(isNumber,desc="font size"),
        fillColor = AttrMapValue(isColorOrNone,desc="color of the font"),
        textAnchor = AttrMapValue(OneOf('start','middle','end','numeric'),desc="treat (x,y) as one of the options below."),
        encoding = AttrMapValue(isString),
        textRenderMode = AttrMapValue(OneOf(0,1,2,3,4,5,6,7),desc="Control whether text is filled/stroked etc etc"),
        )
    encoding = 'utf8'

    def __init__(self, x, y, text, **kw):
        self.x = x
        self.y = y
        self.text = text
        self.textAnchor = 'start'
        self.fontName = STATE_DEFAULTS['fontName']
        self.fontSize = STATE_DEFAULTS['fontSize']
        self.fillColor = STATE_DEFAULTS['fillColor']
        self.setProperties(kw)

    def getEast(self):
        return self.x + stringWidth(self.text,self.fontName,self.fontSize, self.encoding)

    def copy(self):
        new = self.__class__(self.x, self.y, self.text)
        new.setProperties(self.getProperties())
        return new

    def getBounds(self):
        # assumes constant drop of 0.2*size to baseline
        t = self.text
        w = stringWidth(t,self.fontName,self.fontSize,self.encoding)
        tA = self.textAnchor
        x = self.x
        if tA!='start':
            if tA=='middle':
                x -= 0.5*w
            elif tA=='end':
                x -= w
            elif tA=='numeric':
                x -= numericXShift(tA,t,w,self.fontName,self.fontSize,self.encoding)
        return (x, self.y - 0.2 * self.fontSize, x+w, self.y + self.fontSize)

class UserNode(_DrawTimeResizeable):
    """A simple template for creating a new node.  The user (Python
    programmer) may subclasses this.  provideNode() must be defined to
    provide a Shape primitive when called by a renderer.  It does
    NOT inherit from Shape, as the renderer always replaces it, and
    your own classes can safely inherit from it without getting
    lots of unintended behaviour."""

    def provideNode(self):
        """Override this to create your own node. This lets widgets be
        added to drawings; they must create a shape (typically a group)
        so that the renderer can draw the custom node."""

        raise NotImplementedError("this method must be redefined by the user/programmer")

class DirectDraw(Shape):
    """try to draw directly on the canvas"""
    def drawDirectly(self,canvas):
        raise NotImplementedError("this method must be redefined by the user/programmer")

def test():
    r = Rect(10,10,200,50)
    import pprint
    pp = pprint.pprint
    w = sys.stdout.write
    w('a Rectangle: ')
    pp(r.getProperties())
    w('\nverifying...')
    r.verify()
    w(' OK\n')
    #print 'setting rect.z = "spam"'
    #r.z = 'spam'
    w('deleting rect.width ')
    del r.width
    w('verifying...')
    r.verify()

if __name__=='__main__':
    test()
